from ..models.enums import ViralLoads


class DataRegistry:
    """Registry to hold data values."""

    version = None

    expiration_particle = {
        "particle_size_range": {
            "long_range": {
                "minimum_diameter": 0.1,
                "maximum_diameter": 30,
                "references": "Morawska et al. (https://doi.org/10.1016/j.jaerosci.2008.11.002); Johnson et al. (https://doi.org/10.1016/j.jaerosci.2011.07.009).",
            },
            "short_range": {
                "minimum_diameter": 0.1,
                "maximum_diameter": 100,
                "references": "Morawska et al. (https://doi.org/10.1016/j.jaerosci.2008.11.002); Johnson et al. (https://doi.org/10.1016/j.jaerosci.2011.07.009).",
            },
        },   
        "BLOmodel": {
            "cn": {"B": 0.06, "L": 0.2, "O": 0.0010008},
            "mu": {"B": 0.989541, "L": 1.38629, "O": 4.97673},
            "sigma": {"B": 0.262364, "L": 0.506818, "O": 0.585005},
            "references": "Morawska et al. (https://doi.org/10.1016/j.jaerosci.2008.11.002); Johnson et al. (https://doi.org/10.1016/j.jaerosci.2011.07.009).",
        },
        "expiration_BLO_factors": {
            "Breathing": {"B": 1., "L": 0., "O": 0., },
            "Speaking": {"B": 1., "L": 1., "O": 1., },
            "Singing": {"B": 1., "L": 5., "O": 5., },
            "Shouting": {"B": 1., "L": 5., "O": 5., },
            "references": "Morawska et al. (https://doi.org/10.1016/j.jaerosci.2008.11.002); Johnson et al. (https://doi.org/10.1016/j.jaerosci.2011.07.009).",
        },
        "particle": {
            "evaporation_factor": 0.3,
            "references": "Marr et al. (https://doi.org/10.1098/rsif.2018.0298).",
        },
    }

    activity_distributions = {
        "Seated": {
            "inhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": -0.6872121723362303,
                    "lognormal_standard_deviation_gaussian": 0.10498338229297108,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
            "exhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": -0.6872121723362303,
                    "lognormal_standard_deviation_gaussian": 0.10498338229297108,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
        },
        "Standing": {
            "inhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": -0.5742377578494785,
                    "lognormal_standard_deviation_gaussian": 0.09373162411398223,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
            "exhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": -0.5742377578494785,
                    "lognormal_standard_deviation_gaussian": 0.09373162411398223,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
        },
        "Light activity": {
            "inhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": 0.21380242785625422,
                    "lognormal_standard_deviation_gaussian": 0.09435378091059601,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
            "exhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": 0.21380242785625422,
                    "lognormal_standard_deviation_gaussian": 0.09435378091059601,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
        },
        "Moderate activity": {
            "inhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": 0.551771330362601,
                    "lognormal_standard_deviation_gaussian": 0.1894616357138137,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
            "exhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": 0.551771330362601,
                    "lognormal_standard_deviation_gaussian": 0.1894616357138137,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
        },
        "Heavy exercise": {
            "inhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": 1.1644665696723049,
                    "lognormal_standard_deviation_gaussian": 0.21744554768657565,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
            "exhalation_rate": {
                "associated_value": "Log-normal distribution",
                "parameters": {
                    "lognormal_mean_gaussian": 1.1644665696723049,
                    "lognormal_standard_deviation_gaussian": 0.21744554768657565,
                },
                "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
            },
        },
    }
    
    virological_data = {
        "symptomatic_vl_frequencies": {
            "associated_value": "Log Custom Kernel distribution",
            "parameters": {
                "log_variable": [
                    2.46032,
                    2.67431,
                    2.85434,
                    3.06155,
                    3.25856,
                    3.47256,
                    3.66957,
                    3.85979,
                    4.09927,
                    4.27081,
                    4.47631,
                    4.66653,
                    4.87204,
                    5.10302,
                    5.27456,
                    5.46478,
                    5.6533,
                    5.88428,
                    6.07281,
                    6.30549,
                    6.48552,
                    6.64856,
                    6.85407,
                    7.10373,
                    7.30075,
                    7.47229,
                    7.66081,
                    7.85782,
                    8.05653,
                    8.27053,
                    8.48453,
                    8.65607,
                    8.90573,
                    9.06878,
                    9.27429,
                    9.473,
                    9.66152,
                    9.87552,
                ],
                "frequencies": [
                    0.001206885,
                    0.007851618,
                    0.008078144,
                    0.01502491,
                    0.013258014,
                    0.018528495,
                    0.020053765,
                    0.021896167,
                    0.022047184,
                    0.018604005,
                    0.01547796,
                    0.018075445,
                    0.021503523,
                    0.022349217,
                    0.025097721,
                    0.032875078,
                    0.030594727,
                    0.032573045,
                    0.034717482,
                    0.034792991,
                    0.033267721,
                    0.042887485,
                    0.036846816,
                    0.03876473,
                    0.045016819,
                    0.040063473,
                    0.04883754,
                    0.043944602,
                    0.048142864,
                    0.041588741,
                    0.048762031,
                    0.027921732,
                    0.033871788,
                    0.022122693,
                    0.016927718,
                    0.008833228,
                    0.00478598,
                    0.002807662,
                ],
                "kernel_bandwidth": 0.1,
            },
            "references": "Henriques et al. (https://doi.org/10.1101/2021.10.14.21264988) and references therein.",
        },
        "covid_overal_vl_data": {
            "associated_value": "Weibull distribution",
            "parameters": {
                "shape_factor": 3.47,
                "scale_factor": 7.01,
                "start": 0.01,
                "stop": 0.99,
                "num": 30.0,
                "min_bound": 2,
                "max_bound": 10,
                "interpolation_fp_left": 0,
                "interpolation_fp_right": 0,
                "max_function": 0.2,
            },
            "references": "Chen et al. (https://elifesciences.org/articles/65774); First line of the figure in https://iiif.elifesciences.org/lax:65774%2Felife-65774-fig4-figsupp3-v2.tif/full/1500,/0/default.jpg.",
        },
        "virus_distributions": {
            "SARS_CoV_2": {
                "viral_load_in_sputum": ViralLoads.COVID_OVERALL.value,
                "infectious_dose": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 10, "high": 100},
                    "references": "Lednicky et al. (https://doi.org/10.1016/j.ijid.2020.09.025); Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "viable_to_RNA_ratio": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 0.01, "high": 0.6},
                    "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "transmissibility_factor": {
                    "value": 1,
                    "references": "Campbell et al. (https://doi.org/10.2807/1560-7917.ES.2021.26.24.2100509.)",
                },
                "infectiousness_days": {
                    "value": 14,
                    "references": "From French and Swiss guidelines (2021), supported by WHO (https://iris.who.int/bitstream/handle/10665/342004/WHO-2019-nCoV-IHR-Quarantine-2021.1-eng.pdf). Only used for incidence rate calculations.",
                },
            },
            "SARS_CoV_2_ALPHA": {
                "viral_load_in_sputum": ViralLoads.COVID_OVERALL.value,
                "infectious_dose": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 10, "high": 100},
                    "references": "Lednicky et al. (https://doi.org/10.1016/j.ijid.2020.09.025); Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "viable_to_RNA_ratio": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 0.01, "high": 0.6},
                    "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "transmissibility_factor": {
                    "value": 0.78,
                    "references": "Campbell et al. (https://doi.org/10.2807/1560-7917.ES.2021.26.24.2100509.)",
                },
                "infectiousness_days": {
                    "value": 14,
                    "references": "From French and Swiss guidelines (2021), supported by WHO (https://iris.who.int/bitstream/handle/10665/342004/WHO-2019-nCoV-IHR-Quarantine-2021.1-eng.pdf). Only used for incidence rate calculations.",
                },
            },
            "SARS_CoV_2_BETA": {
                "viral_load_in_sputum": ViralLoads.COVID_OVERALL.value,
                "infectious_dose": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 10, "high": 100},
                    "references": "Lednicky et al. (https://doi.org/10.1016/j.ijid.2020.09.025); Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "viable_to_RNA_ratio": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 0.01, "high": 0.6},
                    "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "transmissibility_factor": {
                    "value": 0.8,
                    "references": "Campbell et al. (https://doi.org/10.2807/1560-7917.ES.2021.26.24.2100509.)",
                },
                "infectiousness_days": {
                    "value": 14,
                    "references": "From French and Swiss guidelines (2021), supported by WHO (https://iris.who.int/bitstream/handle/10665/342004/WHO-2019-nCoV-IHR-Quarantine-2021.1-eng.pdf). Only used for incidence rate calculations.",
                },
            },
            "SARS_CoV_2_GAMMA": {
                "viral_load_in_sputum": ViralLoads.COVID_OVERALL.value,
                "infectious_dose": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 10, "high": 100},
                    "references": "Lednicky et al. (https://doi.org/10.1016/j.ijid.2020.09.025); Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "viable_to_RNA_ratio": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 0.01, "high": 0.6},
                    "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "transmissibility_factor": {
                    "value": 0.72,
                    "references": "Campbell et al. (https://doi.org/10.2807/1560-7917.ES.2021.26.24.2100509.)",
                },
                "infectiousness_days": {
                    "value": 14,
                    "references": "From French and Swiss guidelines (2021), supported by WHO (https://iris.who.int/bitstream/handle/10665/342004/WHO-2019-nCoV-IHR-Quarantine-2021.1-eng.pdf). Only used for incidence rate calculations.",
                },
            },
            "SARS_CoV_2_DELTA": {
                "viral_load_in_sputum": ViralLoads.COVID_OVERALL.value,
                "infectious_dose": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 10, "high": 100},
                    "references": "Lednicky et al. (https://doi.org/10.1016/j.ijid.2020.09.025); Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "viable_to_RNA_ratio": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 0.01, "high": 0.6},
                    "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "transmissibility_factor": {
                    "value": 0.51,
                    "references": "Campbell et al. (https://doi.org/10.2807/1560-7917.ES.2021.26.24.2100509.)",
                },
                "infectiousness_days": {
                    "value": 14,
                    "references": "From French and Swiss guidelines (2021), supported by WHO (https://iris.who.int/bitstream/handle/10665/342004/WHO-2019-nCoV-IHR-Quarantine-2021.1-eng.pdf). Only used for incidence rate calculations.",
                },
            },
            "SARS_CoV_2_OMICRON": {
                "viral_load_in_sputum": ViralLoads.COVID_OVERALL.value,
                "infectious_dose": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 10, "high": 100},
                    "references": "Lednicky et al. (https://doi.org/10.1016/j.ijid.2020.09.025); Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "viable_to_RNA_ratio": {
                    "associated_value": "Uniform distribution",
                    "parameters": {"low": 0.01, "high": 0.6},
                    "references": "Henriques et al. (https://doi.org/10.1098/rsfs.2021.0076) and references therein.",
                },
                "transmissibility_factor": {
                    "value": 0.2,
                    "references": "Du et al. (https://www.mdpi.com/1999-4915/14/4/821).",
                },
                "infectiousness_days": {
                    "value": 14,
                    "references": "From French and Swiss guidelines (2021), supported by WHO (https://iris.who.int/bitstream/handle/10665/342004/WHO-2019-nCoV-IHR-Quarantine-2021.1-eng.pdf). Only used for incidence rate calculations.",
                },
            },
        },
    }

    mask_distributions = {
        "Type I": {
            "η_inhale": {
                "associated_value": "Uniform distribution",
                "parameters": {
                    "low": 0.25,
                    "high": 0.80,
                },
                "references": "Pan et al. (https://doi.org/10.1080/02786826.2021.1890687); Booth et al. (https://doi.org/10.1016/j.jhin.2013.02.007); Monn et al. (https://doi.org/10.4209/aaqr.2020.08.0531).",
            },
            "factor_exhale": {
                "value": 1,
                "references": "",
            },
        },
        "FFP2": {
            "η_inhale": {
                "associated_value": "Uniform distribution",
                "parameters": {
                    "low": 0.83,
                    "high": 0.91,
                },
                "references": "Pan et al. (https://doi.org/10.1080/02786826.2021.1890687); Booth et al. (https://doi.org/10.1016/j.jhin.2013.02.007); Monn et al. (https://doi.org/10.4209/aaqr.2020.08.0531).",
            },
            "factor_exhale": {
                "value": 1,
                "references": "",
            },
        },
        "Cloth": {
            "η_inhale": {
                "associated_value": "Uniform distribution",
                "parameters": {
                    "low": 0.05,
                    "high": 0.40,
                },
                "references": "Pan et al. (https://doi.org/10.1080/02786826.2021.1890687); Booth et al. (https://doi.org/10.1016/j.jhin.2013.02.007); Monn et al. (https://doi.org/10.4209/aaqr.2020.08.0531).",
            },
            "η_exhale": {
                "associated_value": "Uniform distribution",
                "parameters": {
                    "low": 0.20,
                    "high": 0.50,
                },
                "references": "Pan et al. (https://doi.org/10.1080/02786826.2021.1890687); Booth et al. (https://doi.org/10.1016/j.jhin.2013.02.007); Monn et al. (https://doi.org/10.4209/aaqr.2020.08.0531).",
            },
            "factor_exhale": {
                "value": 1,
                "references": "",
            },
        },
    }

    ####################################

    room = {
        "inside_temp": 293.,
        "humidity_with_heating": 0.3,
        "humidity_without_heating": 0.5,
        "references": "N/A.",
    }

    ventilation = {
        "natural": {
            "discharge_factor": {
                "sliding": 0.6,
            },
        },
        "infiltration_ventilation": 0.25,
        "references": "Henriques et al. (https://doi.org/10.1101/2021.10.14.21264988).",
    }

    concentration_model = {
        "virus_concentration_model": {
            "min_background_concentration": 0.0,
            "references": "N/A.",
        },
        "CO2_concentration_model": {
            "CO2_atmosphere_concentration": 440.44,
            "CO2_fraction_exhaled": 0.042,
            "references": "ECMWF - https://charts.ecmwf.int/products/carbon-dioxide-forecasts; Zhang et al. (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7127751/).",
        },
    }

    short_range_model = {
        "dilution_factor": {
            "mouth_diameter": 0.02,
            "exhalation_coefficient": 2,
            "breathing_cycle": 4,
            "penetration_coefficients": {
                "𝛽r1": 0.18,
                "𝛽r2": 0.2,
                "𝛽x1": 2.4,
            },
            "references": "Jia et al. (https://doi.org/10.1016/j.buildenv.2022.109166).",
        },
        "conversational_distance": {
            "minimum_distance": 0.5,
            "maximum_distance": 2.0,
            "references": "Derived from Fig. 8 a) 'stand-stand' in Zhang et al. (https://www.mdpi.com/1660-4601/17/4/1445).",
        },
    }

    monte_carlo = {
        "sample_size": 250000,
        "references": "N/A.",
    }

    population_scenario_activity = {
        "office": {"placeholder": "Office", "activity": "Seated", "expiration": {"Speaking": 1, "Breathing": 2}, "references": "N/A."},
        "smallmeeting": {
            "placeholder": "Small meeting (<10 occ.)",
            "activity": "Seated",
            "expiration": {"Speaking": 1},
            "references": "N/A.",
        },
        "largemeeting": {
            "placeholder": "Large meeting (>= 10 occ.)",
            "activity": "Standing",
            "expiration": {"Speaking": 1, "Breathing": 2},
            "references": "N/A.",
        },
        "callcenter": {"placeholder": "Call Center", "activity": "Seated", "expiration": {"Speaking": 1}},
        "controlroom-day": {
            "placeholder": "Control Room - Day shift",
            "activity": "Seated",
            "expiration": {"Speaking": 1, "Breathing": 1},
            "references": "N/A.",
        },
        "controlroom-night": {
            "placeholder": "Control Room - Night shift",
            "activity": "Seated",
            "expiration": {"Speaking": 1, "Breathing": 9},
            "references": "N/A.",
        },
        "library": {"placeholder": "Library", "activity": "Seated", "expiration": {"Breathing": 1}, "references": "N/A."},
        "lab": {
            "placeholder": "Lab",
            "activity": "Light activity",
            "expiration": {"Speaking": 1, "Breathing": 1},
            "references": "N/A.",
        },
        "workshop": {
            "placeholder": "Workshop",
            "activity": "Moderate activity",
            "expiration": {"Speaking": 1, "Breathing": 1},
            "references": "N/A.",
        },
        "training": {"placeholder": "Conference/Training (speaker infected)", "activity": "Standing", "expiration": {"Speaking": 1}, "references": "N/A."},
        "training_attendee": {"placeholder": "Conference/Training (attendee infected)", "activity": "Seated", "expiration": {"Breathing": 1}, "references": "N/A."},
        "gym": {"placeholder": "Gym", "activity": "Heavy exercise", "expiration": {"Breathing": 1}, "references": "N/A."},
        "household-day": {
            "placeholder": "Household (day time)",
            "activity": "Light activity",
            "expiration": {"Breathing": 5, "Speaking": 5},
            "references": "N/A.",
        },
        "household-night": {
            "placeholder": "Household (evening and night time)",
            "activity": "Seated",
            "expiration": {"Breathing": 7, "Speaking": 3},
            "references": "N/A."
        },
        "primary-school": {
            "placeholder": "Primary school",
            "activity": "Light activity",
            "expiration": {"Breathing": 5, "Speaking": 5},
            "references": "N/A.",
        },
        "secondary-school": {
            "placeholder": "Secondary school",
            "activity": "Light activity",
            "expiration": {"Breathing": 7, "Speaking": 3},
            "references": "N/A.",
        },
        "university": {
            "placeholder": "University",
            "activity": "Seated",
            "expiration": {"Breathing": 9, "Speaking": 1},
            "references": "N/A.",
        },
        "restaurant": {
            "placeholder": "Restaurant",
            "activity": "Seated",
            "expiration": {"Breathing": 1, "Speaking": 9},
            "references": "N/A.",
        },
        "precise": {"placeholder": "Precise", "activity": "", "expiration": {}, "references": "N/A."},
    }

    def to_dict(self):
        # Filter out methods, special attributes, and non-serializable objects
        data_dict = {
            key: value
            for key, value in self.__class__.__dict__.items()
            if not key.startswith("__") and not callable(value) and not isinstance(value, (type, classmethod, staticmethod))
        }
        return data_dict

    def update(self, data, version=None):
        """Update local cache with data provided as argument."""
        for attr_name, value in data.items():
            setattr(self, attr_name, value)

        if version:
            self.version = version



import dataclasses
import datetime
import logging
import typing
import re

import numpy as np

from caimira import __version__ as calculator_version
from ..form_validator import FormData, cast_class_fields, time_string_to_minutes
from ..defaults import (DEFAULTS, CONFIDENCE_LEVEL_OPTIONS,
                        MECHANICAL_VENTILATION_TYPES, MASK_WEARING_OPTIONS, MONTH_NAMES, VACCINE_BOOSTER_TYPE, VACCINE_TYPE,
                        VENTILATION_TYPES, VOLUME_TYPES, WINDOWS_OPENING_REGIMES, WINDOWS_TYPES)
from ...models import models, data, monte_carlo as mc
from ...models.monte_carlo.data import activity_distributions, virus_distributions, mask_distributions, short_range_distances
from ...models.monte_carlo.data import expiration_distribution, expiration_BLO_factors, expiration_distributions, short_range_expiration_distributions

LOG = logging.getLogger("MODEL")

minutes_since_midnight = typing.NewType('minutes_since_midnight', int)


@dataclasses.dataclass
class VirusFormData(FormData):
    ascertainment_bias: str
    activity_type: str
    air_changes: float
    air_supply: float
    arve_sensors_option: bool
    calculator_version: str
    ceiling_height: float
    CO2_fitting_result: dict
    conditional_probability_viral_loads: bool
    event_month: str
    exposure_option: str
    floor_area: float
    geographic_cases: int
    geographic_population: int
    hepa_amount: float
    hepa_option: bool
    humidity: str
    inside_temp: float
    location_latitude: float
    location_longitude: float
    location_name: str
    mask_type: str
    mask_wearing_option: str
    mechanical_ventilation_type: str
    opening_distance: float
    precise_activity: dict
    room_heating_option: bool
    room_number: str
    sensor_in_use: str
    short_range_interactions: list
    short_range_occupants: int
    short_range_option: str
    simulation_name: str
    vaccine_booster_option: bool
    vaccine_booster_type: str
    vaccine_option: bool
    vaccine_type: str
    ventilation_type: str
    virus_type: str
    volume_type: str
    window_height: float
    window_opening_regime: str
    window_type: str
    window_width: float
    windows_duration: float
    windows_frequency: float
    windows_number: int

    _DEFAULTS: typing.ClassVar[typing.Dict[str, typing.Any]] = DEFAULTS

    def validate(self):
        # Validate population parameters 
        self.validate_population_parameters()

        validation_tuples = [('activity_type', self.data_registry.population_scenario_activity.keys()),
                             ('ascertainment_bias', CONFIDENCE_LEVEL_OPTIONS),
                             ('event_month', MONTH_NAMES),
                             ('mechanical_ventilation_type', MECHANICAL_VENTILATION_TYPES),
                             ('mask_type', list(mask_distributions(self.data_registry).keys())),
                             ('mask_wearing_option', MASK_WEARING_OPTIONS),
                             ('vaccine_type', VACCINE_TYPE),
                             ('vaccine_booster_type', VACCINE_BOOSTER_TYPE),
                             ('ventilation_type', VENTILATION_TYPES),
                             ('virus_type', list(virus_distributions(self.data_registry).keys())),
                             ('volume_type', VOLUME_TYPES),
                             ('window_opening_regime', WINDOWS_OPENING_REGIMES),
                             ('window_type', WINDOWS_TYPES)]

        for attr_name, valid_set in validation_tuples:
            if getattr(self, attr_name) not in valid_set:
                raise ValueError(
                    f"{getattr(self, attr_name)} is not a valid value for {attr_name}")

        # Validate number of infected people == 1 when activity is Conference/Training.
        if self.activity_type == 'training' and self.infected_people > 1:
            raise ValueError(
                'Conference/Training activities are limited to 1 infected.')

        # Validate ventilation parameters
        if self.ventilation_type == 'natural_ventilation':
            if self.window_type == 'not-applicable':
                raise ValueError(
                    "window_type cannot be 'not-applicable' if "
                    "ventilation_type is 'natural_ventilation'"
                )
            if self.window_opening_regime == 'not-applicable':
                raise ValueError(
                    "window_opening_regime cannot be 'not-applicable' if "
                    "ventilation_type is 'natural_ventilation'"
                )
            if (self.window_opening_regime == 'windows_open_periodically' and
                    self.windows_duration > self.windows_frequency):
                raise ValueError(
                    'Duration cannot be bigger than frequency.'
                )

        if (self.ventilation_type == 'mechanical_ventilation'
                and self.mechanical_ventilation_type == 'not-applicable'):
            raise ValueError("mechanical_ventilation_type cannot be 'not-applicable' if "
                             "ventilation_type is 'mechanical_ventilation'")

        # Validate specific inputs - breaks (exposed and infected)
        if self.specific_breaks != {}:
            if type(self.specific_breaks) is not dict:
                raise TypeError(
                    'The specific breaks should be in a dictionary.')

            dict_keys = list(self.specific_breaks.keys())
            if "exposed_breaks" not in dict_keys:
                raise TypeError(
                    f'Unable to fetch "exposed_breaks" key. Got "{dict_keys[0]}".')
            if "infected_breaks" not in dict_keys:
                raise TypeError(
                    f'Unable to fetch "infected_breaks" key. Got "{dict_keys[1]}".')

            for population_breaks in ['exposed_breaks', 'infected_breaks']:
                if self.specific_breaks[population_breaks] != []:
                    if type(self.specific_breaks[population_breaks]) is not list:
                        raise TypeError(
                            f'All breaks should be in a list. Got {type(self.specific_breaks[population_breaks])}.')
                    for input_break in self.specific_breaks[population_breaks]:
                        # Input validations.
                        if type(input_break) is not dict:
                            raise TypeError(
                                f'Each break should be a dictionary. Got {type(input_break)}.')
                        dict_keys = list(input_break.keys())
                        if "start_time" not in input_break:
                            raise TypeError(
                                f'Unable to fetch "start_time" key. Got "{dict_keys[0]}".')
                        if "finish_time" not in input_break:
                            raise TypeError(
                                f'Unable to fetch "finish_time" key. Got "{dict_keys[1]}".')
                        for time in input_break.values():
                            if not re.compile("^(2[0-3]|[01]?[0-9]):([0-5]?[0-9])$").match(time):
                                raise TypeError(
                                    f'Wrong time format - "HH:MM". Got "{time}".')

        # Validate specific inputs - precise activity
        if self.precise_activity != {}:
            if type(self.precise_activity) is not dict:
                raise TypeError(
                    'The precise activities should be in a dictionary.')

            dict_keys = list(self.precise_activity.keys())
            if "physical_activity" not in dict_keys:
                raise TypeError(
                    f'Unable to fetch "physical_activity" key. Got "{dict_keys[0]}".')
            if "respiratory_activity" not in dict_keys:
                raise TypeError(
                    f'Unable to fetch "respiratory_activity" key. Got "{dict_keys[1]}".')

            if type(self.precise_activity['physical_activity']) is not str:
                raise TypeError(
                    'The physical activities should be a single string.')

            if type(self.precise_activity['respiratory_activity']) is not list:
                raise TypeError(
                    'The respiratory activities should be in a list.')

            total_percentage = 0
            for respiratory_activity in self.precise_activity['respiratory_activity']:
                if type(respiratory_activity) is not dict:
                    raise TypeError(
                        'Each respiratory activity should be defined in a dictionary.')
                dict_keys = list(respiratory_activity.keys())
                if "type" not in dict_keys:
                    raise TypeError(
                        f'Unable to fetch "type" key. Got "{dict_keys[0]}".')
                if "percentage" not in dict_keys:
                    raise TypeError(
                        f'Unable to fetch "percentage" key. Got "{dict_keys[1]}".')
                total_percentage += respiratory_activity['percentage']

            if total_percentage != 100:
                raise ValueError(
                    f'The sum of all respiratory activities should be 100. Got {total_percentage}.')

        # Validate number of people with short-range interactions
        if self.occupancy_format == "static": max_occupants_for_sr = self.total_people - self.infected_people
        else: max_occupants_for_sr = np.max(np.array([entry["total_people"] for entry in self.dynamic_exposed_occupancy]))
        if self.short_range_occupants > max_occupants_for_sr:
            raise ValueError(
                f'The total number of occupants having short-range interactions ({self.short_range_occupants}) should be lower than the exposed population ({max_occupants_for_sr}).'
            )
        
        # Validate short-range interactions interval
        if self.short_range_option == "short_range_yes":
            for interaction in self.short_range_interactions:
                # Check if presence is within long-range exposure
                presence = self.short_range_interval(interaction)
                if (self.occupancy_format == 'dynamic'):
                    long_range_start = min(time_string_to_minutes(self.dynamic_infected_occupancy[0]['start_time']), 
                                            time_string_to_minutes(self.dynamic_exposed_occupancy[0]['start_time']))
                    long_range_stop = max(time_string_to_minutes(self.dynamic_infected_occupancy[-1]['finish_time']), 
                                            time_string_to_minutes(self.dynamic_exposed_occupancy[-1]['finish_time']))
                else:
                    long_range_start = min(self.infected_start, self.exposed_start)
                    long_range_stop = max(self.infected_finish, self.exposed_finish)
                if not (long_range_start/60 <= presence.present_times[0][0] <= long_range_stop/60 and 
                        long_range_start/60 <= presence.present_times[0][-1] <= long_range_stop/60):
                    raise ValueError(f"Short-range interactions should be defined during simulation time. Got {interaction}")

    def initialize_room(self) -> models.Room:
        # Initializes room with volume either given directly or as product of area and height
        if self.volume_type == 'room_volume_explicit':
            volume = self.room_volume
        else:
            volume = self.floor_area * self.ceiling_height

        if self.arve_sensors_option == False:
            if self.room_heating_option:
                humidity = self.data_registry.room['humidity_with_heating']
            else:
                humidity = self.data_registry.room['humidity_without_heating']
            inside_temp = self.data_registry.room['inside_temp']
        else:
            humidity = float(self.humidity)
            inside_temp = self.inside_temp

        return models.Room(volume=volume, inside_temp=models.PiecewiseConstant((0, 24), (inside_temp,)), humidity=humidity) # type: ignore

    def build_mc_model(self) -> mc.ExposureModel:
        room = self.initialize_room()
        ventilation: models._VentilationBase = self.ventilation()
        infected_population: models.InfectedPopulation = self.infected_population()
        short_range = []
        if self.short_range_option == "short_range_yes":
            for interaction in self.short_range_interactions:
                short_range.append(mc.ShortRangeModel(
                    data_registry=self.data_registry,
                    expiration=short_range_expiration_distributions(
                        self.data_registry)[interaction['expiration']],
                    activity=infected_population.activity,
                    presence=self.short_range_interval(interaction),
                    distance=short_range_distances(self.data_registry),
                ))

        return mc.ExposureModel(
            data_registry=self.data_registry,
            concentration_model=mc.ConcentrationModel(
                data_registry=self.data_registry,
                room=room,
                ventilation=ventilation,
                infected=infected_population,
                evaporation_factor=0.3,
            ),
            short_range=tuple(short_range),
            exposed=self.exposed_population(),
            geographical_data=mc.Cases(
                geographic_population=self.geographic_population,
                geographic_cases=self.geographic_cases,
                ascertainment_bias=CONFIDENCE_LEVEL_OPTIONS[self.ascertainment_bias],
            ),
            exposed_to_short_range=self.short_range_occupants,
        )

    def build_model(self, sample_size=None) -> models.ExposureModel:
        sample_size = sample_size or self.data_registry.monte_carlo['sample_size']
        return self.build_mc_model().build_model(size=sample_size)

    def build_CO2_model(self, sample_size=None) -> models.CO2ConcentrationModel:
        sample_size = sample_size or self.data_registry.monte_carlo['sample_size']
        infected_population: models.InfectedPopulation = self.infected_population(
        ).build_model(sample_size)
        exposed_population: models.Population = self.exposed_population().build_model(sample_size)

        state_change_times = set(
            infected_population.presence_interval().transition_times())
        state_change_times.update(
            exposed_population.presence_interval().transition_times())
        transition_times = sorted(state_change_times)

        total_people = [infected_population.people_present(stop) + exposed_population.people_present(stop)
                        for _, stop in zip(transition_times[:-1], transition_times[1:])]

        if (self.activity_type == 'precise'):
            activity_defn, _ = self.generate_precise_activity_expiration()
        else:
            activity_defn = self.data_registry.population_scenario_activity[
                self.activity_type]['activity']

        population = mc.SimplePopulation(
            number=models.IntPiecewiseConstant(transition_times=tuple(
                transition_times), values=tuple(total_people)),
            presence=None,
            activity=activity_distributions(self.data_registry)[activity_defn],
        )

        # Builds a CO2 concentration model based on model inputs
        return mc.CO2ConcentrationModel(
            data_registry=self.data_registry,
            room=self.initialize_room(),
            ventilation=self.ventilation(),
            CO2_emitters=population,
        ).build_model(size=sample_size)

    def tz_name_and_utc_offset(self) -> typing.Tuple[str, float]:
        """
        Return the timezone name (e.g. CET), and offset, in hours, that need to
        be *added* to UTC to convert to the form location's timezone.

        """
        month = MONTH_NAMES.index(self.event_month) + 1
        timezone = data.weather.timezone_at(
            latitude=self.location_latitude, longitude=self.location_longitude,
        )
        # We choose the first of the month for the current year.
        date = datetime.datetime(datetime.datetime.now().year, month, 1)
        name = timezone.tzname(date)
        assert isinstance(name, str)
        utc_offset_td = timezone.utcoffset(date)
        assert isinstance(utc_offset_td, datetime.timedelta)
        utc_offset_hours = utc_offset_td.total_seconds() / 60 / 60
        return name, utc_offset_hours

    def outside_temp(self) -> models.PiecewiseConstant:
        """
        Return the outside temperature as a PiecewiseConstant in the destination
        timezone.

        """
        month = MONTH_NAMES.index(self.event_month) + 1

        wx_station = self.nearest_weather_station()
        temp_profile = data.weather.mean_hourly_temperatures(
            wx_station=wx_station[0], month=MONTH_NAMES.index(self.event_month) + 1)

        _, utc_offset = self.tz_name_and_utc_offset()

        # Offset the source times according to the difference from UTC (as a
        # result the first data value may no longer be a midnight, and the hours
        # no longer ordered modulo 24).
        source_times = np.arange(24) + utc_offset
        times, temp_profile = data.weather.refine_hourly_data(
            source_times,
            temp_profile,
            npts=24*10,  # 10 steps per hour => 6 min steps
        )
        outside_temp = models.PiecewiseConstant(
            tuple(float(t) for t in times), tuple(float(t)
                                                  for t in temp_profile),
        )
        return outside_temp

    def ventilation(self) -> models._VentilationBase:
        always_on = models.PeriodicInterval(period=120, duration=120)
        periodic_interval = models.PeriodicInterval(self.windows_frequency, self.windows_duration,
                                                    min(self.infected_start, self.exposed_start)/60)
        if self.ventilation_type == 'from_fitting':
            ventilations = []
            transition_times = self.CO2_fitting_result['transition_times']
            for index, (start, stop) in enumerate(zip(transition_times[:-1], transition_times[1:])):
                ventilations.append(models.AirChange(active=models.SpecificInterval(present_times=((start, stop), )),
                                                     air_exch=self.CO2_fitting_result['ventilation_values'][index]))
            return models.MultipleVentilation(tuple(ventilations))

        # Initializes a ventilation instance as a window if 'natural_ventilation' is selected, or as a HEPA-filter otherwise
        if self.ventilation_type == 'natural_ventilation':
            if self.window_opening_regime == 'windows_open_periodically':
                window_interval = periodic_interval
            else:
                window_interval = always_on

            outside_temp = self.outside_temp()

            ventilation: models.Ventilation
            if self.window_type == 'window_sliding':
                ventilation = models.SlidingWindow(
                    data_registry=self.data_registry,
                    active=window_interval,
                    outside_temp=outside_temp,
                    window_height=self.window_height,
                    opening_length=self.opening_distance,
                    number_of_windows=self.windows_number,
                )
            elif self.window_type == 'window_hinged':
                ventilation = models.HingedWindow(
                    active=window_interval,
                    outside_temp=outside_temp,
                    window_height=self.window_height,
                    window_width=self.window_width,
                    opening_length=self.opening_distance,
                    number_of_windows=self.windows_number,
                )

        elif self.ventilation_type == "no_ventilation":
            ventilation = models.AirChange(active=always_on, air_exch=0.)
        else:
            if self.mechanical_ventilation_type == 'mech_type_air_changes':
                ventilation = models.AirChange(
                    active=always_on, air_exch=self.air_changes)
            else:
                ventilation = models.HVACMechanical(
                    active=always_on, q_air_mech=self.air_supply)

        # This is a minimal, always present source of ventilation, due
        # to the air infiltration from the outside.
        # See CERN-OPEN-2021-004, p. 12.
        residual_vent: float = self.data_registry.ventilation['infiltration_ventilation'] # type: ignore
        infiltration_ventilation = models.AirChange(
            active=always_on, air_exch=residual_vent)
        if self.hepa_option:
            hepa = models.HEPAFilter(
                active=always_on, q_air_mech=self.hepa_amount)
            return models.MultipleVentilation((ventilation, hepa, infiltration_ventilation))
        else:
            return models.MultipleVentilation((ventilation, infiltration_ventilation))

    def nearest_weather_station(self) -> data.weather.WxStationRecordType:
        """Return the nearest weather station (which has valid data) for this form"""
        return data.weather.nearest_wx_station(
            longitude=self.location_longitude, latitude=self.location_latitude
        )

    def mask(self) -> models.Mask:
        # Initializes the mask type if mask wearing is "continuous", otherwise instantiates the mask attribute as
        # the "No mask"-mask
        if self.mask_wearing_option == 'mask_on':
            mask = mask_distributions(self.data_registry)[self.mask_type]
        else:
            mask = models.Mask.types['No mask']
        return mask

    def generate_precise_activity_expiration(self) -> typing.Tuple[typing.Any, ...]:
        # It means the precise activity is not defined by a specific input.
        if self.precise_activity == {}:
            return ()
        respiratory_dict = {}
        for respiratory_activity in self.precise_activity['respiratory_activity']:
            respiratory_dict[respiratory_activity['type']
                             ] = respiratory_activity['percentage']

        return (self.precise_activity['physical_activity'], respiratory_dict)

    def infected_population(self) -> mc.InfectedPopulation:
        # Initializes the virus
        virus = virus_distributions(self.data_registry)[self.virus_type]

        # Occupancy
        if self.occupancy_format == 'dynamic':
            if isinstance(self.dynamic_infected_occupancy, typing.List) and len(self.dynamic_infected_occupancy) > 0:
                # If dynamic occupancy is defined, the generator will parse and validate the
                # respective input to a format readable by the model - `IntPiecewiseConstant`.
                infected_occupancy = self.generate_dynamic_occupancy(self.dynamic_infected_occupancy)
                infected_presence = None
            else:
                raise TypeError(f'If dynamic occupancy is selected, a populated list of occupancy intervals is expected. Got "{self.dynamic_infected_occupancy}".')
        else:
            # The number of exposed occupants is the total number of occupants
            # minus the number of infected occupants.
            infected_occupancy = self.infected_people
            infected_presence = self.infected_present_interval()

        # Activity and expiration
        activity_defn = self.data_registry.population_scenario_activity[self.activity_type]['activity']
        expiration_defn = self.data_registry.population_scenario_activity[self.activity_type]['expiration']
        if (self.activity_type == 'smallmeeting'):
            # Conversation of N people is approximately 1/N% of the time speaking.
            total_people: int = max(infected_occupancy.values) if self.occupancy_format == 'dynamic' else self.total_people
            expiration_defn = {'Speaking': 1, 'Breathing': total_people - 1}
        elif (self.activity_type == 'precise'):
            activity_defn, expiration_defn = self.generate_precise_activity_expiration()

        activity = activity_distributions(self.data_registry)[activity_defn]
        expiration = build_expiration(self.data_registry, expiration_defn)

        infected = mc.InfectedPopulation(
            data_registry=self.data_registry,
            number=infected_occupancy,
            presence=infected_presence,
            virus=virus,
            mask=self.mask(),
            activity=activity,
            expiration=expiration,
            # Vaccination status does not affect the infected population (for now)
            host_immunity=0.,
        )
        return infected

    def exposed_population(self) -> mc.Population:
        activity_defn = (self.precise_activity['physical_activity']
                         if self.activity_type == 'precise'
                         else str(self.data_registry.population_scenario_activity[self.activity_type]['activity']))
        activity = activity_distributions(self.data_registry)[activity_defn]

        if self.occupancy_format == 'dynamic':
            if isinstance(self.dynamic_exposed_occupancy, typing.List) and len(self.dynamic_exposed_occupancy) > 0:
                # If dynamic occupancy is defined, the generator will parse and validate the
                # respective input to a format readable by the model - IntPiecewiseConstant.
                exposed_occupancy = self.generate_dynamic_occupancy(self.dynamic_exposed_occupancy)
                exposed_presence = None
            else:
                raise TypeError(f'If dynamic occupancy is selected, a populated list of occupancy intervals is expected. Got "{self.dynamic_exposed_occupancy}".')
        else:
            # The number of exposed occupants is the total number of occupants
            # minus the number of infected occupants.
            exposed_occupancy = self.total_people - self.infected_people
            exposed_presence = self.exposed_present_interval()

        if (self.vaccine_option):
            if (self.vaccine_booster_option and self.vaccine_booster_type != 'Other'):
                host_immunity = [vaccine['VE'] for vaccine in data.vaccine_booster_host_immunity if
                                 vaccine['primary series vaccine'] == self.vaccine_type and
                                 vaccine['booster vaccine'] == self.vaccine_booster_type][0]
            else:
                host_immunity = data.vaccine_primary_host_immunity[self.vaccine_type]
        else:
            host_immunity = 0.

        exposed = mc.Population(
            number=exposed_occupancy,
            presence=exposed_presence,
            activity=activity,
            mask=self.mask(),
            host_immunity=host_immunity,
        )
        return exposed

    def short_range_interval(self, interaction) -> models.SpecificInterval:
        start_time = time_string_to_minutes(interaction['start_time'])
        duration = float(interaction['duration'])
        return models.SpecificInterval(present_times=((start_time/60, (start_time + duration)/60),))


def build_expiration(data_registry, expiration_definition) -> mc._ExpirationBase:
    if isinstance(expiration_definition, str):
        return expiration_distributions(data_registry)[expiration_definition]
    elif isinstance(expiration_definition, dict):
        total_weight = sum(expiration_definition.values())
        BLO_factors = np.sum([
            np.array(expiration_BLO_factors(data_registry)
                     [exp_type]) * weight/total_weight
            for exp_type, weight in expiration_definition.items()
        ], axis=0)
        return expiration_distribution(data_registry=data_registry, BLO_factors=tuple(BLO_factors))


def baseline_raw_form_data() -> typing.Dict[str, typing.Union[str, float]]:
    # Note: This isn't a special "baseline". It can be updated as required.
    return {
        'activity_type': 'office',
        'air_changes': '',
        'air_supply': '',
        'ascertainment_bias': 'confidence_low',
        'calculator_version': calculator_version,
        'ceiling_height': '',
        'conditional_probability_viral_loads': '0',
        'dynamic_exposed_occupancy': '[]',
        'dynamic_infected_occupancy': '[]',
        'event_month': 'January',
        'exposed_coffee_break_option': 'coffee_break_4',
        'exposed_coffee_duration': '10',
        'exposed_finish': '18:00',
        'exposed_lunch_finish': '13:30',
        'exposed_lunch_option': '1',
        'exposed_lunch_start': '12:30',
        'exposed_start': '09:00',
        'floor_area': '',
        'geographic_cases': 0,
        'geographic_population': 0,
        'hepa_amount': '250',
        'hepa_option': '0',
        'humidity': '0.5',
        'infected_coffee_break_option': 'coffee_break_4',
        'infected_coffee_duration': '10',
        'infected_dont_have_breaks_with_exposed': '1',
        'infected_finish': '18:00',
        'infected_lunch_finish': '13:30',
        'infected_lunch_option': '1',
        'infected_lunch_start': '12:30',
        'infected_people': '1',
        'infected_start': '09:00',
        'inside_temp': '293.',
        'location_latitude': 46.20833,
        'location_longitude': 6.14275,
        'location_name': 'Geneva',
        'mask_type': 'Type I',
        'mask_wearing_option': 'mask_off',
        'mechanical_ventilation_type': '',
        'occupancy_format': 'static',
        'opening_distance': '0.2',
        'room_heating_option': '0',
        'room_number': '123',
        'room_volume': '75',
        'short_range_interactions': '[]',
        'short_range_option': 'short_range_no',
        'simulation_name': 'Test',
        'total_people': '10',
        'vaccine_booster_option': '0',
        'vaccine_booster_type': 'AZD1222_(AstraZeneca)',
        'vaccine_option': '0',
        'vaccine_type': 'Ad26.COV2.S_(Janssen)',
        'ventilation_type': 'natural_ventilation',
        'virus_type': 'SARS_CoV_2',
        'volume_type': 'room_volume_explicit',
        'window_height': '2',
        'window_opening_regime': 'windows_open_permanently',
        'window_type': 'window_sliding',
        'window_width': '2',
        'windows_duration': '10',
        'windows_frequency': '60',
        'windows_number': '1',
    }

cast_class_fields(VirusFormData)



import dataclasses
import logging
import typing
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
import pandas as pd
import re

from ..form_validator import FormData, cast_class_fields
from ..defaults import NO_DEFAULT
from ...store.data_registry import DataRegistry
from ...models import models
from ...report.virus_report_data import img2base64, _figure2bytes

minutes_since_midnight = typing.NewType('minutes_since_midnight', int)

LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class CO2FormData(FormData):
    CO2_data: dict
    fitting_ventilation_states: list
    room_capacity: typing.Optional[int]

    # The default values for undefined fields. Note that the defaults here
    # and the defaults in any html form must not be contradictory.
    _DEFAULTS: typing.ClassVar[typing.Dict[str, typing.Any]] = {
        'CO2_data': '{}',
        'dynamic_infected_occupancy': '[]',
        'dynamic_exposed_occupancy': '[]',
        'exposed_coffee_break_option': 'coffee_break_0',
        'exposed_coffee_duration': 5,
        'exposed_finish': '17:30',
        'exposed_lunch_finish': '13:30',
        'exposed_lunch_option': True,
        'exposed_lunch_start': '12:30',
        'exposed_start': '08:30',
        'fitting_ventilation_states': '[]',
        'infected_coffee_break_option': 'coffee_break_0',
        'infected_coffee_duration': 5,
        'infected_dont_have_breaks_with_exposed': False,
        'infected_finish': '17:30',
        'infected_lunch_finish': '13:30',
        'infected_lunch_option': True,
        'infected_lunch_start': '12:30',
        'infected_people': 1,
        'infected_start': '08:30',
        'occupancy_format': 'static',
        'room_capacity': None,
        'room_volume': NO_DEFAULT,
        'specific_breaks': '{}',
        'total_people': NO_DEFAULT,
    }

    def __init__(self, **kwargs):
        # Set default values defined in CO2FormData
        for key, value in self._DEFAULTS.items():
            setattr(self, key, kwargs.get(key, value))

        # Initialize the Data Registry
        self.data_registry = DataRegistry()

    def validate(self):
        # Validate population parameters 
        self.validate_population_parameters()

        # Validate room capacity
        if self.room_capacity:
            if type(self.room_capacity) is not int:
                raise TypeError(f'The room capacity should be a valid integer (> 0). Got {type(self.room_capacity)}.')
            if self.room_capacity <= 0:
                raise TypeError(f'The room capacity should be a valid integer (> 0). Got {self.room_capacity}.')

        # Validate specific inputs - breaks (exposed and infected)
        if self.specific_breaks != {}:
            if type(self.specific_breaks) is not dict:
                raise TypeError('The specific breaks should be in a dictionary.')

            dict_keys = list(self.specific_breaks.keys())
            if "exposed_breaks" not in dict_keys:
                raise TypeError(f'Unable to fetch "exposed_breaks" key. Got "{dict_keys[0]}".')
            if "infected_breaks" not in dict_keys:
                raise TypeError(f'Unable to fetch "infected_breaks" key. Got "{dict_keys[1]}".')

            for population_breaks in ['exposed_breaks', 'infected_breaks']:
                if self.specific_breaks[population_breaks] != []:
                    if type(self.specific_breaks[population_breaks]) is not list:
                        raise TypeError(f'All breaks should be in a list. Got {type(self.specific_breaks[population_breaks])}.')
                    for input_break in self.specific_breaks[population_breaks]:
                        # Input validations.
                        if type(input_break) is not dict:
                            raise TypeError(f'Each break should be a dictionary. Got {type(input_break)}.')
                        dict_keys = list(input_break.keys())
                        if "start_time" not in input_break:
                            raise TypeError(f'Unable to fetch "start_time" key. Got "{dict_keys[0]}".')
                        if "finish_time" not in input_break:
                            raise TypeError(f'Unable to fetch "finish_time" key. Got "{dict_keys[1]}".')
                        for time in input_break.values():
                            if not re.compile("^(2[0-3]|[01]?[0-9]):([0-5]?[0-9])$").match(time):
                                raise TypeError(f'Wrong time format - "HH:MM". Got "{time}".')        

    def find_change_points(self) -> list:
        """
        Perform change point detection using scipy library (find_peaks method) with rolling average of data.
        Incorporate existing state change candidates and adjust the result accordingly.
        Returns a list of the detected ventilation transition times, discarding any occupancy state change.
        """
        times: list = self.CO2_data['times']
        CO2_values: list = self.CO2_data['CO2']

        if len(times) != len(CO2_values):
            raise ValueError("times and CO2 values must have the same length.")

        # Time difference between two consecutive time data entries, in seconds
        diff = (times[1] - times[0]) * 3600 # Initial data points in absolute hours, e.g. 14.78

        # Calculate minimum interval for smoothing technique
        smooth_min_interval_in_minutes = 1 # Minimum time difference for smooth technique
        window_size = max(int((smooth_min_interval_in_minutes * 60) // diff), 1)

        # Applying a rolling average to smooth the initial data
        smoothed_co2 = pd.Series(CO2_values).rolling(window=window_size, center=True).mean()

        # Calculate minimum interval for peaks and valleys detection
        peak_valley_min_interval_in_minutes = 15 # Minimum time difference between two peaks or two valleys
        min_distance_points = max(int((peak_valley_min_interval_in_minutes * 60) // diff), 1)

        # Calculate minimum width of datapoints for valley detection
        width_min_interval_in_minutes = 20 # Minimum duration of a valley
        min_valley_width = max(int((width_min_interval_in_minutes * 60) // diff), 1)

        # Find peaks (maxima) in the smoothed data applying the distance factor
        peaks, _ = find_peaks(smoothed_co2.values, prominence=100, distance=min_distance_points)
        
        # Find valleys (minima) by inverting the smoothed data and applying the width and distance factors
        valleys, _ = find_peaks(-smoothed_co2.values, prominence=50, width=min_valley_width, distance=min_distance_points)

        # Extract peak and valley timestamps
        timestamps = np.array(times)
        peak_timestamps = timestamps[peaks]
        valley_timestamps = timestamps[valleys]

        return sorted(np.concatenate((peak_timestamps, valley_timestamps)))

    def generate_ventilation_plot(self,
                                  ventilation_transition_times: typing.Optional[list] = None,
                                  occupancy_transition_times: typing.Optional[list] = None,
                                  predictive_CO2: typing.Optional[list] = None):
            
            # Plot data (x-axis: times; y-axis: CO2 concentrations)
            times_values: list = self.CO2_data['times']
            CO2_values: list = self.CO2_data['CO2']

            fig = plt.figure(figsize=(7, 4), dpi=110)
            plt.plot(times_values, CO2_values, label='CO₂ Data')
            
            # Add predictive CO2
            if (predictive_CO2):
                plt.plot(times_values, predictive_CO2, label='Predictive CO₂')

            # Add ventilation transition times:
            if (ventilation_transition_times):
                for i, time in enumerate(ventilation_transition_times):
                    if i == 0:
                        label = 'Ventilation transition times (suggestion)' if occupancy_transition_times else 'Ventilation transition times'
                    else: label = None
                    plt.axvline(x = time, color = 'red', linewidth=1, linestyle='--', label=label)
            
            # Add occupancy changes (UI):
            if (occupancy_transition_times):
                for i, time in enumerate(occupancy_transition_times):
                    plt.axvline(x = time, color = 'grey', linewidth=1, linestyle='--', label='Occupancy change (from UI)' if i == 0 else None)

            plt.xlabel('Time of day')
            plt.ylabel('Concentration (ppm)')
            plt.legend()

            vent_plot_data = {
                'plot': img2base64(_figure2bytes(fig)),
                'times': times_values,
                'CO2': CO2_values,
                'occ_trans_time': occupancy_transition_times,
                'vent_trans_time': ventilation_transition_times,
                'predictive_CO2': predictive_CO2,
            }

            return img2base64(_figure2bytes(fig)), vent_plot_data

    def population_present_changes(self, infected_presence: models.Interval, exposed_presence: models.Interval) -> typing.List[float]:
        state_change_times = set(infected_presence.transition_times())
        state_change_times.update(exposed_presence.transition_times())
        return sorted(state_change_times)

    def ventilation_transition_times(self) -> typing.Tuple[float]:
        '''
        Check if the last time from the input data is
        included in the ventilation ventilations state.
        Given that the last time is a required state change, 
        if not included, this method adds it.
        '''
        vent_states = self.fitting_ventilation_states
        last_time_from_input = self.CO2_data['times'][-1]
        if (vent_states and last_time_from_input != vent_states[-1]): # The last time value is always needed for the last ACH interval.
            vent_states.append(last_time_from_input)
        return tuple(vent_states)

    def build_model(self, sample_size = None) -> models.CO2DataModel:
        # Build a simple infected and exposed population for the case when presence
        # intervals and number of people are dynamic. Activity type is not needed.
        if self.occupancy_format == 'dynamic':
            if isinstance(self.dynamic_infected_occupancy, typing.List) and len(self.dynamic_infected_occupancy) > 0:
                infected_people = self.generate_dynamic_occupancy(self.dynamic_infected_occupancy)
                infected_presence = None
            else:
                raise TypeError(f'If dynamic occupancy is selected, a populated list of occupancy intervals is expected. Got "{self.dynamic_infected_occupancy}".')
            if isinstance(self.dynamic_exposed_occupancy, typing.List) and len(self.dynamic_exposed_occupancy) > 0:
                exposed_people = self.generate_dynamic_occupancy(self.dynamic_exposed_occupancy)
                exposed_presence = None
            else:
                raise TypeError(f'If dynamic occupancy is selected, a populated list of occupancy intervals is expected. Got "{self.dynamic_exposed_occupancy}".')
        else:
            infected_people = self.infected_people
            exposed_people = self.total_people - self.infected_people
            infected_presence = self.infected_present_interval()
            exposed_presence = self.exposed_present_interval()

        infected_population = models.SimplePopulation(
            number=infected_people,
            presence=infected_presence,
            activity=None, # type: ignore
        )
        exposed_population=models.SimplePopulation(
            number=exposed_people,
            presence=exposed_presence,
            activity=None, # type: ignore
        )
        
        all_state_changes=self.population_present_changes(infected_population.presence_interval(), 
                                                          exposed_population.presence_interval())
        total_people = [infected_population.people_present(stop) + exposed_population.people_present(stop)
                        for _, stop in zip(all_state_changes[:-1], all_state_changes[1:])]

        return models.CO2DataModel(
                data_registry=self.data_registry,
                room=models.Room(volume=self.room_volume, capacity=self.room_capacity),
                occupancy=models.IntPiecewiseConstant(transition_times=tuple(all_state_changes), values=tuple(total_people)),
                ventilation_transition_times=self.ventilation_transition_times(),
                times=self.CO2_data['times'],
                CO2_concentrations=self.CO2_data['CO2'],
            )

cast_class_fields(CO2FormData)



import typing

from caimira.calculator.validators.co2.co2_validator import CO2FormData
from caimira.calculator.store.data_registry import DataRegistry
from caimira.calculator.models.models import CO2DataModel


def generate_form_obj(form_data: typing.Dict, data_registry: DataRegistry) -> CO2FormData:
    return CO2FormData.from_dict(form_data=form_data, data_registry=data_registry)


def generate_model(form_obj: CO2FormData) -> CO2DataModel:
    return form_obj.build_model()


def generate_report(model: CO2DataModel) -> typing.Dict:
    return dict(model.CO2_fit_params())


def request_CO2_transition_times(form_data: dict) -> dict:
    """
    Calculate and return the transition times related to CO2 levels including
    the ventilation transition times (identified from the change point algorithm)
    and relevant occupancy transition times (first and last occurrences).
    """
    data_registry = DataRegistry()

    form_obj = generate_form_obj(form_data=form_data, data_registry=data_registry)
    CO2model = generate_model(form_obj=form_obj)
    
    # Occupancy transition times
    occupancy_transition_times = list(CO2model.occupancy.transition_times)
    relevant_occupancy_times = [occupancy_transition_times[0]] + [occupancy_transition_times[-1]]
    # Ventilation transition times
    vent_transition_times = form_obj.find_change_points()
    ventilation_times = sorted(vent_transition_times)
    # Total transition times
    total_times = sorted(relevant_occupancy_times + ventilation_times)
    
    return {
        "occupancy_times": occupancy_transition_times,
        "ventilation_times": ventilation_times,
        "total_times": total_times
    }


def request_CO2_report(form_data: typing.Dict) -> typing.Dict:
    data_registry: DataRegistry = DataRegistry()

    form_obj: CO2FormData = generate_form_obj(form_data=form_data, data_registry=data_registry)
    model: CO2DataModel = generate_model(form_obj=form_obj)
    report_data: typing.Dict = generate_report(model=model)

    return report_data



from enum import Enum

class ViralLoads(Enum):
    COVID_OVERALL = "Ref: Viral load - covid overal viral load data"
    SYMPTOMATIC_FREQUENCIES = "Ref: Viral load - symptomatic viral load frequencies"



import os
import tornado.web


class CorsHandler(tornado.web.RequestHandler):
    """Handler to manage CORS (Cross-Origin Resource Sharing) configuration.

    This handler implements CORS support by setting appropriate headers for cross-origin requests.
    It checks allowed origins against environment variable CAIMIRA_ALLOWED_ORIGINS and enables
    CORS for matching origins.
    """

    def set_default_headers(self):
        allowed_origins = os.environ.get("CAIMIRA_ALLOWED_ORIGINS", None)
        request_origin = self.request.headers.get("Origin", None)  # can have value `null`
        if allowed_origins and request_origin:
            allowed_origins = [origin.lower().strip() for origin in allowed_origins.split(",")]
            if request_origin.lower() in allowed_origins:
                self.set_header("Access-Control-Allow-Origin", request_origin)

    def options(self, *args):
        self.set_header("Access-Control-Allow-Methods", "POST, OPTIONS")
        self.set_header("Access-Control-Allow-Headers", "Content-Type")
        self.set_status(204)  # No Content


class BaseRequestHandler(CorsHandler):
    """Base handler for HTTP requests extending CorsHandler.

    This class provides base functionality for handling HTTP requests with CORS support.
    It includes error handling capabilities by overriding the write_error method.
    """

    def write_error(self, status_code, **kwargs):
        self.set_status(status_code)
        self.write({"message": kwargs.get('exc_info')[1].__str__()})



from caimira.api.routes.base_handler import BaseRequestHandler

from caimira import __version__


class LandingPageHandler(BaseRequestHandler):
    def get(self):
        self.write("<h1>Welcome to the CAiMIRA REST API</h1>")
        self.write(f"<h4>Code Version: {__version__}</h4>")
        self.write("<h4>Official GitLab: <a href='https://gitlab.cern.ch/caimira/caimira' target='_blank'>https://gitlab.cern.ch/caimira/caimira</a></h4>")
        self.write("<h4>Documentation: <a href='https://caimira.docs.cern.ch/' target='_blank'>https://caimira.docs.cern.ch/</a><h4>")



import concurrent.futures
import base64
import dataclasses
import io
import typing
import numpy as np
import matplotlib.pyplot as plt

from caimira.calculator.models import models, dataclass_utils, profiler, monte_carlo as mc
from caimira.calculator.models.enums import ViralLoads
from caimira.calculator.validators.virus.virus_validator import VirusFormData


def model_start_end(model: models.ExposureModel):
    t_start = min(model.exposed.presence_interval().boundaries()[0][0],
                  model.concentration_model.infected.presence_interval().boundaries()[0][0])
    t_end = max(model.exposed.presence_interval().boundaries()[-1][1],
                model.concentration_model.infected.presence_interval().boundaries()[-1][1])
    return t_start, t_end


def fill_big_gaps(array, gap_size):
    """
    Insert values into the given sorted list if there is a gap of more than ``gap_size``.
    All values in the given array are preserved, even if they are within the ``gap_size`` of one another.

    >>> fill_big_gaps([1, 2, 4], gap_size=0.75)
    [1, 1.75, 2, 2.75, 3.5, 4]

    """
    result = []
    if len(array) == 0:
        raise ValueError("Input array must be len > 0")

    last_value = array[0]
    for value in array:
        while value - last_value > gap_size + 1e-15:
            last_value = last_value + gap_size
            result.append(last_value)
        result.append(value)
        last_value = value
    return result


def non_temp_transition_times(model: models.ExposureModel):
    """
    Return the non-temperature (and PiecewiseConstant) based transition times.

    """
    def walk_model(model, name=""):
        # Extend walk_dataclass to handle lists of dataclasses
        # (e.g. in MultipleVentilation).
        for name, obj in dataclass_utils.walk_dataclass(model, name=name):
            if name.endswith('.ventilations') and isinstance(obj, (list, tuple)):
                for i, item in enumerate(obj):
                    fq_name_i = f'{name}[{i}]'
                    yield fq_name_i, item
                    if dataclasses.is_dataclass(item):
                        yield from dataclass_utils.walk_dataclass(item, name=fq_name_i)
            else:
                yield name, obj

    t_start, t_end = model_start_end(model)

    change_times = {t_start, t_end}
    for name, obj in walk_model(model, name="exposure"):
        if isinstance(obj, models.Interval):
            change_times |= obj.transition_times()

    # Only choose times that are in the range of the model (removes things
    # such as PeriodicIntervals, which extend beyond the model itself).
    return sorted(time for time in change_times if (t_start <= time <= t_end))


def interesting_times(model: models.ExposureModel, approx_n_pts: typing.Optional[int] = None) -> typing.List[float]:
    """
    Pick approximately ``approx_n_pts`` time points which are interesting for the
    given model. If not provided by argument, ``approx_n_pts`` is set to be 15 times
    the number of hours of the simulation.

    Initially the times are seeded by important state change times (excluding
    outside temperature), and the times are then subsequently expanded to ensure
    that the step size is at most ``(t_end - t_start) / approx_n_pts``.

    """
    times = non_temp_transition_times(model)
    sim_duration = max(times) - min(times)
    if not approx_n_pts:
        approx_n_pts = sim_duration * 15

    # Expand the times list to ensure that we have a maximum gap size between
    # the key times.
    nice_times = fill_big_gaps(times, gap_size=(sim_duration) / approx_n_pts)
    return nice_times


def concentrations_with_sr_breathing(form: VirusFormData, model: models.ExposureModel, times: typing.List[float], short_range_intervals: typing.List) -> typing.List[float]:
    lower_concentrations = []
    for time in times:
        for index, (start, stop) in enumerate(short_range_intervals):
            # For visualization issues, add short-range breathing activity to the initial long-range concentrations
            if start <= time <= stop and form.short_range_interactions[index]['expiration'] == 'Breathing':
                lower_concentrations.append(
                    np.array(model.concentration(float(time))).mean())
                break
        lower_concentrations.append(
            np.array(model.concentration_model.concentration(float(time))).mean())
    return lower_concentrations


def _calculate_deposited_exposure(model, time1, time2, fn_name=None):
    return np.array(model.deposited_exposure_between_bounds(float(time1), float(time2))).mean(), fn_name


def _calculate_long_range_deposited_exposure(model, time1, time2, fn_name=None):
    return np.array(model.long_range_deposited_exposure_between_bounds(float(time1), float(time2))).mean(), fn_name


def _calculate_co2_concentration(CO2_model, time, fn_name=None):
    return np.array(CO2_model.concentration(float(time))).mean(), fn_name


@profiler.profile
def calculate_report_data(form: VirusFormData, executor_factory: typing.Callable[[], concurrent.futures.Executor]) -> typing.Dict[str, typing.Any]:
    model: models.ExposureModel = form.build_model()
    
    times = interesting_times(model)
    short_range_intervals = [interaction.presence.boundaries()[0]
                             for interaction in model.short_range]
    short_range_expirations = [interaction['expiration']
                               for interaction in form.short_range_interactions] if form.short_range_option == "short_range_yes" else []

    concentrations = [
        np.array(model.concentration(float(time))).mean()
        for time in times
    ]
    lower_concentrations = concentrations_with_sr_breathing(
        form, model, times, short_range_intervals)

    CO2_model: models.CO2ConcentrationModel = form.build_CO2_model()

    # compute deposited exposures and CO2 concentrations in parallel to increase performance
    deposited_exposures = []
    long_range_deposited_exposures = []
    CO2_concentrations = []

    tasks = []
    with executor_factory() as executor:
        for time1, time2 in zip(times[:-1], times[1:]):
            tasks.append(executor.submit(
                _calculate_deposited_exposure, model, time1, time2, fn_name="de"))
            tasks.append(executor.submit(
                _calculate_long_range_deposited_exposure, model, time1, time2, fn_name="lr"))
            # co2 concentration: takes each time as param, not the interval
            tasks.append(executor.submit(
                _calculate_co2_concentration, CO2_model, time1, fn_name="co2"))
        # co2 concentration: calculate the last time too
        tasks.append(executor.submit(_calculate_co2_concentration,
                     CO2_model, times[-1], fn_name="co2"))

    for task in tasks:
        result, fn_name = task.result()
        if fn_name == "de":
            deposited_exposures.append(result)
        elif fn_name == "lr":
            long_range_deposited_exposures.append(result)
        elif fn_name == "co2":
            CO2_concentrations.append(result)

    cumulative_doses = np.cumsum(deposited_exposures)
    long_range_cumulative_doses = np.cumsum(long_range_deposited_exposures)

    prob = np.array(model.infection_probability())
    prob_dist_count, prob_dist_bins = np.histogram(prob/100, bins=100, density=True)
    
    # Probabilistic exposure and expected new cases (only for static occupancy)
    prob_probabilistic_exposure = None
    expected_new_cases = None
    if form.occupancy_format == "static":
        if form.exposure_option == "p_probabilistic_exposure":
            prob_probabilistic_exposure = np.array(model.total_probability_rule()).mean()
        expected_new_cases =  np.array(model.expected_new_cases()).mean()

    exposed_presence_intervals = [list(interval) for interval in model.exposed.presence_interval().boundaries()]

    conditional_probability_data = None
    uncertainties_plot_src = None
    if (form.conditional_probability_viral_loads and
            model.data_registry.virological_data['virus_distributions'][form.virus_type]['viral_load_in_sputum'] == ViralLoads.COVID_OVERALL.value):  # type: ignore
        # Generate all the required data for the conditional probability plot
        conditional_probability_data = manufacture_conditional_probability_data(
            model, prob)
        # Generate the matplotlib image based on the received data
        uncertainties_plot_src = img2base64(_figure2bytes(
            uncertainties_plot(prob, conditional_probability_data)))

    return {
        "model": model,
        "times": list(times),
        "exposed_presence_intervals": exposed_presence_intervals,
        "short_range_intervals": short_range_intervals,
        "short_range_expirations": short_range_expirations,
        "concentrations": concentrations,
        "concentrations_zoomed": lower_concentrations,
        "cumulative_doses": list(cumulative_doses),
        "long_range_cumulative_doses": list(long_range_cumulative_doses),
        "prob_inf": prob.mean(),
        "prob_inf_sd": prob.std(),
        "prob_dist": list(prob),
        "prob_hist_count": list(prob_dist_count),
        "prob_hist_bins": list(prob_dist_bins),
        "prob_probabilistic_exposure": prob_probabilistic_exposure,
        "expected_new_cases": expected_new_cases,
        "uncertainties_plot_src": uncertainties_plot_src,
        "CO2_concentrations": CO2_concentrations,
        "conditional_probability_data": conditional_probability_data,
    }


def conditional_prob_inf_given_vl_dist(
    infection_probability: models._VectorisedFloat,
    viral_loads: np.ndarray,
    specific_vl: models._VectorisedFloat,
    step: models._VectorisedFloat
):

    pi_means = []
    lower_percentiles = []
    upper_percentiles = []

    for vl_log in viral_loads:
        # Probability of infection corresponding to a certain viral load value in the distribution
        specific_prob = infection_probability[np.where(
            (vl_log-step/2-specific_vl)*(vl_log+step/2-specific_vl) < 0)[0]]  # type: ignore

        pi_means.append(specific_prob.mean())
        lower_percentiles.append(np.quantile(specific_prob, 0.05))
        upper_percentiles.append(np.quantile(specific_prob, 0.95))

    return pi_means, lower_percentiles, upper_percentiles


def manufacture_conditional_probability_data(
    exposure_model: models.ExposureModel,
    infection_probability: models._VectorisedFloat
):
    min_vl = 2
    max_vl = 10
    step = (max_vl - min_vl)/100
    viral_loads = np.arange(min_vl, max_vl, step)
    specific_vl = np.log10(
        exposure_model.concentration_model.virus.viral_load_in_sputum)
    pi_means, lower_percentiles, upper_percentiles = conditional_prob_inf_given_vl_dist(infection_probability, viral_loads,
                                                                                        specific_vl, step)
    log10_vl_in_sputum = np.log10(
        exposure_model.concentration_model.infected.virus.viral_load_in_sputum)

    return {
        'viral_loads': list(viral_loads),
        'pi_means': list(pi_means),
        'lower_percentiles': list(lower_percentiles),
        'upper_percentiles': list(upper_percentiles),
        'log10_vl_in_sputum': list(log10_vl_in_sputum),
    }


def uncertainties_plot(infection_probability: models._VectorisedFloat,
                       conditional_probability_data: dict):

    viral_loads: list = conditional_probability_data['viral_loads']
    pi_means: list = conditional_probability_data['pi_means']
    lower_percentiles: list = conditional_probability_data['lower_percentiles']
    upper_percentiles: list = conditional_probability_data['upper_percentiles']
    log10_vl_in_sputum: list = conditional_probability_data['log10_vl_in_sputum']

    fig, ((axs00, axs01, axs02), (axs10, axs11, axs12)) = plt.subplots(nrows=2, ncols=3,  # type: ignore
                                                                       gridspec_kw={'width_ratios': [5, 0.5] + [1],
                                                                                    'height_ratios': [3, 1], 'wspace': 0},
                                                                       sharey='row',
                                                                       sharex='col')

    axs01.axis('off')
    axs11.axis('off')
    axs12.axis('off')

    axs01.set_visible(False)

    axs00.plot(viral_loads, np.array(pi_means),
               label='Predictive total probability')
    axs00.fill_between(viral_loads, np.array(lower_percentiles), np.array(
        upper_percentiles), alpha=0.1, label='5ᵗʰ and 95ᵗʰ percentile')

    axs02.hist(infection_probability, bins=30, orientation='horizontal')
    axs02.set_xticks([])
    axs02.set_xticklabels([])
    axs02.set_facecolor("lightgrey")

    highest_bar = axs02.get_xlim()[1]
    axs02.set_xlim(0, highest_bar)

    axs02.text(highest_bar * 0.5, 50,
               "$P(I)=$\n" + rf"$\bf{np.round(np.mean(infection_probability), 1)}$%", ha='center', va='center')
    axs10.hist(log10_vl_in_sputum,
               bins=150, range=(2, 10), color='grey')
    axs10.set_facecolor("lightgrey")
    axs10.set_yticks([])
    axs10.set_yticklabels([])
    axs10.set_xticks([i for i in range(2, 13, 2)])
    axs10.set_xticklabels(['$10^{' + str(i) + '}$' for i in range(2, 13, 2)])
    axs10.set_xlim(2, 10)
    axs10.set_xlabel('Viral load\n(RNA copies)', fontsize=12)
    axs00.set_ylabel('Conditional Probability\nof Infection', fontsize=12)

    axs00.text(9.5, -0.01, '$(i)$')
    axs10.text(9.5, axs10.get_ylim()[1] * 0.8, '$(ii)$')
    axs02.set_title('$(iii)$', fontsize=10)

    axs00.legend()
    return fig


def _figure2bytes(figure):
    # Draw the image
    img_data = io.BytesIO()
    figure.savefig(img_data, format='png', bbox_inches="tight",
                   transparent=True, dpi=110)
    return img_data


def img2base64(img_data) -> str:
    img_data.seek(0)
    pic_hash = base64.b64encode(img_data.read()).decode('ascii')
    # A src suitable for a tag such as f'<img id="scenario_concentration_plot" src="{result}">.
    return f'data:image/png;base64,{pic_hash}'


def calculate_vl_scenarios_percentiles(model: mc.ExposureModel) -> typing.Dict[str, mc.ExposureModel]:
    viral_load = model.concentration_model.infected.virus.viral_load_in_sputum
    scenarios = {}
    for percentil in (0.01, 0.05, 0.25, 0.5, 0.75, 0.95, 0.99):
        vl = np.quantile(viral_load, percentil)
        specific_vl_scenario = dataclass_utils.nested_replace(model,
                                                              {'concentration_model.infected.virus.viral_load_in_sputum': vl}
                                                              )
        scenarios[str(vl)] = np.mean(
            specific_vl_scenario.infection_probability())
    return {
        'alternative_viral_load': scenarios,
    }


def manufacture_alternative_scenarios(form: VirusFormData) -> typing.Dict[str, mc.ExposureModel]:
    scenarios = {}
    if (form.short_range_option == "short_range_no"):
        # Two special option cases - HEPA and/or FFP2 masks.
        FFP2_being_worn = bool(form.mask_wearing_option ==
                               'mask_on' and form.mask_type == 'FFP2')
        if FFP2_being_worn and form.hepa_option:
            FFP2andHEPAalternative = dataclass_utils.replace(
                form, mask_type='Type I')
            if not (form.hepa_option and form.mask_wearing_option == 'mask_on' and form.mask_type == 'Type I'):
                scenarios['Base scenario with HEPA filter and Type I masks'] = FFP2andHEPAalternative.build_mc_model()
        if not FFP2_being_worn and form.hepa_option:
            noHEPAalternative = dataclass_utils.replace(form, mask_type='FFP2')
            noHEPAalternative = dataclass_utils.replace(
                noHEPAalternative, mask_wearing_option='mask_on')
            noHEPAalternative = dataclass_utils.replace(
                noHEPAalternative, hepa_option=False)
            if not (not form.hepa_option and FFP2_being_worn):
                scenarios['Base scenario without HEPA filter, with FFP2 masks'] = noHEPAalternative.build_mc_model()

        # The remaining scenarios are based on Type I masks (possibly not worn)
        # and no HEPA filtration.
        form = dataclass_utils.replace(form, mask_type='Type I')
        if form.hepa_option:
            form = dataclass_utils.replace(form, hepa_option=False)

        with_mask = dataclass_utils.replace(
            form, mask_wearing_option='mask_on')
        without_mask = dataclass_utils.replace(
            form, mask_wearing_option='mask_off')

        if form.ventilation_type == 'mechanical_ventilation':
            # scenarios['Mechanical ventilation with Type I masks'] = with_mask.build_mc_model()
            if not (form.mask_wearing_option == 'mask_off'):
                scenarios['Mechanical ventilation without masks'] = without_mask.build_mc_model(
                )

        elif form.ventilation_type == 'natural_ventilation':
            # scenarios['Windows open with Type I masks'] = with_mask.build_mc_model()
            if not (form.mask_wearing_option == 'mask_off'):
                scenarios['Windows open without masks'] = without_mask.build_mc_model()

        # No matter the ventilation scheme, we include scenarios which don't have any ventilation.
        with_mask_no_vent = dataclass_utils.replace(
            with_mask, ventilation_type='no_ventilation')
        without_mask_or_vent = dataclass_utils.replace(
            without_mask, ventilation_type='no_ventilation')

        if not (form.mask_wearing_option == 'mask_on' and form.mask_type == 'Type I' and form.ventilation_type == 'no_ventilation'):
            scenarios['No ventilation with Type I masks'] = with_mask_no_vent.build_mc_model()
        if not (form.mask_wearing_option == 'mask_off' and form.ventilation_type == 'no_ventilation'):
            scenarios['Neither ventilation nor masks'] = without_mask_or_vent.build_mc_model()

    else:
        # When dynamic occupancy is defined, the replace of total people is useless - the expected number of new cases is not calculated.
        if form.occupancy_format == 'static':
            no_short_range_alternative = dataclass_utils.replace(form, short_range_interactions=[], total_people=form.total_people - form.short_range_occupants)
        elif form.occupancy_format == 'dynamic':
            for occ in form.dynamic_exposed_occupancy: # Update the number of exposed people with long-range exposure
                if occ['total_people'] > form.short_range_occupants: occ['total_people'] = max(0, occ['total_people'] - form.short_range_occupants)
            no_short_range_alternative = dataclass_utils.replace(form, short_range_interactions=[], dynamic_exposed_occupancy=form.dynamic_exposed_occupancy)
        
        scenarios['Base scenario without short-range interactions'] = no_short_range_alternative.build_mc_model()

    return scenarios


def scenario_statistics(
    mc_model: mc.ExposureModel,
    sample_times: typing.List[float],
    static_occupancy: bool,
    compute_prob_exposure: bool,
):
    model = mc_model.build_model(
        size=mc_model.data_registry.monte_carlo['sample_size'])

    return {
        'probability_of_infection': np.mean(model.infection_probability()),
        'expected_new_cases': np.mean(model.expected_new_cases()) if static_occupancy else None,
        'concentrations': [
            np.mean(model.concentration(time))
            for time in sample_times
        ],
        'prob_probabilistic_exposure': model.total_probability_rule() if compute_prob_exposure else None,
    }


def comparison_report(
        form: VirusFormData,
        report_data: typing.Dict[str, typing.Any],
        scenarios: typing.Dict[str, mc.ExposureModel],
        executor_factory: typing.Callable[[], concurrent.futures.Executor],
):
    if (form.short_range_option == "short_range_no"):
        statistics = {
            'Current scenario': {
                'probability_of_infection': report_data['prob_inf'],
                'expected_new_cases': report_data['expected_new_cases'],
                'concentrations': report_data['concentrations'],
            }
        }
    else:
        statistics = {}

    static_occupancy = form.occupancy_format == "static"
    compute_prob_exposure = form.short_range_option == "short_range_yes" and form.exposure_option == "p_probabilistic_exposure" and static_occupancy
        
    with executor_factory() as executor:
        results = executor.map(
            scenario_statistics,
            scenarios.values(),
            [report_data['times']] * len(scenarios),
            [static_occupancy] * len(scenarios),
            [compute_prob_exposure] * len(scenarios),
            timeout=60,
        )

    for (name, model), model_stats in zip(scenarios.items(), results):
        statistics[name] = model_stats

    return {
        'stats': statistics,
    }


def alternative_scenarios_data(form: VirusFormData, report_data: typing.Dict[str, typing.Any], executor_factory: typing.Callable[[], concurrent.futures.Executor]) -> typing.Dict[str, typing.Any]:
    alternative_scenarios: typing.Dict[str, typing.Any] = manufacture_alternative_scenarios(form=form)
    return {
        'alternative_scenarios': comparison_report(form=form, report_data=report_data, scenarios=alternative_scenarios, executor_factory=executor_factory)
    }



from __future__ import annotations

from dataclasses import dataclass
import typing

import numpy as np
from scipy import special as sp
from scipy.stats import weibull_min

from ..enums import ViralLoads

import caimira.calculator.models.monte_carlo.models as mc
from caimira.calculator.models.monte_carlo.sampleable import LogCustom, LogNormal, Normal, LogCustomKernel, CustomKernel, Uniform, Custom
from caimira.calculator.store.data_registry import DataRegistry


def evaluate_vl(root: typing.Dict, value: str, data_registry: DataRegistry):
    if root[value] == ViralLoads.COVID_OVERALL.value:
        return covid_overal_vl_data(data_registry)
    elif root[value] == ViralLoads.SYMPTOMATIC_FREQUENCIES.value:
        return symptomatic_vl_frequencies(data_registry)
    elif root[value] == 'Custom':
        return param_evaluation(root, 'Viral load custom')
    else:
        raise ValueError(f"Invalid ViralLoads value {value}")


sqrt2pi = np.sqrt(2.*np.pi)
sqrt2 = np.sqrt(2.)


def custom_value_type_lookup(dict: dict, key_part: str) -> typing.Any:
    """
    Look up a custom value type based on a partial key.

    Args:
        dict (dict): The root to search.
        key_part (str): The value type key to match.

    Returns:
        str: The associated value.
    """
    try:
        for key, value in dict.items():
            if (key_part in key):
                return value['associated_value']
    except KeyError:
        return f"Key '{key_part}' not found."


def evaluate_custom_value_type(value_type: str, params: typing.Dict) -> typing.Any:
    """
    Evaluate a custom value type.

    Args:
        dist (str): The type of value.
        params (Dict): The parameters for the value type.

    Returns:
        Any: The generated value.

    Raises:
        ValueError: If the value type is not recognized.

    """
    if value_type == 'Constant value':
        return params
    elif value_type == 'Normal distribution':
        return Normal(
            mean=params['normal_mean_gaussian'], 
            standard_deviation=params['normal_standard_deviation_gaussian']
        )
    elif value_type == 'Log-normal distribution':
        return LogNormal(
            mean_gaussian=params['lognormal_mean_gaussian'], 
            standard_deviation_gaussian=params['lognormal_standard_deviation_gaussian']
        )
    elif value_type == 'Uniform distribution':
        return Uniform(
            low=params['low'], 
            high=params['high']
        )
    elif value_type == 'Log Custom Kernel distribution':
        return LogCustomKernel(
            log_variable=np.array(params['log_variable']), 
            frequencies=np.array(params['frequencies']), 
            kernel_bandwidth=params['kernel_bandwidth']
        )
    else:
        raise ValueError('Bad request - value type not found.')


def param_evaluation(root: typing.Dict, param: typing.Union[str, typing.Any]) -> typing.Any:
    """
    Evaluate a parameter from a nested dictionary.

    Args:
        root (dict): The root dictionary.
        param (Union[str, Any]): The parameter to evaluate.

    Returns:
        Any: The evaluated value.

    Raises:
        TypeError: If the type of the parameter is not defined.

    """
    value = root.get(param)

    if isinstance(value, dict):
        value_type: str = root[param]['associated_value']
        params: typing.Dict = root[param]['parameters']
        return evaluate_custom_value_type(value_type, params)

    elif isinstance(value, float) or isinstance(value, int):
        return value

    else:
        raise TypeError('Bad request - type not allowed.')


@dataclass(frozen=True)
class BLOmodel:
    """
    Represents the probability distribution from the BLO model.
    It is a sum of three lognormal distributions, each of the form
    A * cn * (1 / d) * (1 / (np.sqrt(2 * np.pi) * sigma)) *
            np.exp(-(np.log(d)-mu) ** 2 / (2 * sigma ** 2))
    with A the factor in front of the B, L or O mode.
    From G. Johnson et al., Modality of human
    expired aerosol size distributions, Journal of Aerosol Science,
    vol. 42, no. 12, pp. 839 – 851, 2011,
    https://doi.org/10.1016/j.jaerosci.2011.07.009).
    """
    data_registry: DataRegistry
    #: Factors assigned to resp. the B, L and O modes. They are
    # charateristics of the kind of expiratory activity (e.g. breathing,
    # speaking, singing, or shouting). These are applied on top of the
    # cn concentrations (see below), and depend on the kind of activity
    # (breathing, speaking, singing/shouting)
    BLO_factors: typing.Tuple[float, float, float]

    #: cn (cm^-3) for resp. the B, L and O modes. Corresponds to the
    # total concentration of aerosols for each mode.
    @property
    def cn(self) -> typing.Tuple[float, float, float]:
        _cn = self.data_registry.expiration_particle['BLOmodel']['cn'] # type: ignore
        return (_cn['B'],_cn['L'],_cn['O'])

    # Mean of the underlying normal distributions (represents the log of a
    # diameter in microns), for resp. the B, L and O modes.
    @property
    def mu(self) -> typing.Tuple[float, float, float]:
        _mu = self.data_registry.expiration_particle['BLOmodel']['mu'] # type: ignore
        return (_mu['B'], _mu['L'], _mu['O'])

    # Std deviation of the underlying normal distribution, for resp.
    # the B, L and O modes.
    @property
    def sigma(self) -> typing.Tuple[float, float, float]:
        _sigma = self.data_registry.expiration_particle['BLOmodel']['sigma'] # type: ignore
        return (_sigma['B'],_sigma['L'],_sigma['O'])

    def distribution(self, d):
        """
        Returns the raw value of the probability distribution for a
        given diameter d (microns).
        """
        return sum((1 / d) * (A * cn / (sqrt2pi * sigma)) *
                   np.exp(-(np.log(d) - mu) ** 2 / (2 * sigma ** 2))
                   for A, cn, mu, sigma in zip(self.BLO_factors, self.cn,
                                               self.mu, self.sigma))

    def integrate(self, dmin, dmax):
        """
        Returns the integral between dmin and dmax (in microns) of the
        probability distribution.
        """
        result = 0.
        for A, cn, mu, sigma in zip(self.BLO_factors, self.cn, self.mu, self.sigma):
            ymin = (np.log(dmin)-mu)/(sqrt2*sigma)
            ymax = (np.log(dmax)-mu)/(sqrt2*sigma)
            result += A * cn * (sp.erf(ymax)-sp.erf(ymin)) / 2.
        return result


# From https://doi.org/10.1101/2021.10.14.21264988 and references therein
def activity_distributions(data_registry):
    return {
        'Seated': mc.Activity(
            inhalation_rate=param_evaluation(
                data_registry.activity_distributions['Seated'], 'inhalation_rate'),
            exhalation_rate=param_evaluation(
                data_registry.activity_distributions['Seated'], 'exhalation_rate'),
        ),

        'Standing': mc.Activity(
            inhalation_rate=param_evaluation(
                data_registry.activity_distributions['Standing'], 'inhalation_rate'),
            exhalation_rate=param_evaluation(
                data_registry.activity_distributions['Standing'], 'exhalation_rate'),
        ),

        'Light activity': mc.Activity(
            inhalation_rate=param_evaluation(
                data_registry.activity_distributions['Light activity'], 'inhalation_rate'),
            exhalation_rate=param_evaluation(
                data_registry.activity_distributions['Light activity'], 'exhalation_rate'),
        ),

        'Moderate activity': mc.Activity(
            inhalation_rate=param_evaluation(
                data_registry.activity_distributions['Moderate activity'], 'inhalation_rate'),
            exhalation_rate=param_evaluation(
                data_registry.activity_distributions['Moderate activity'], 'exhalation_rate'),
        ),

        'Heavy exercise': mc.Activity(
            inhalation_rate=param_evaluation(
                data_registry.activity_distributions['Heavy exercise'], 'inhalation_rate'),
            exhalation_rate=param_evaluation(
                data_registry.activity_distributions['Heavy exercise'], 'exhalation_rate'),
        ),
    }


# From https://doi.org/10.1098/rsfs.2021.0076 and references therein
def symptomatic_vl_frequencies(data_registry):
    return param_evaluation(data_registry.virological_data, 'symptomatic_vl_frequencies')


# Weibull distribution with a shape factor of 3.47 and a scale factor of 7.01.
# From https://elifesciences.org/articles/65774 and first line of the figure in
# https://iiif.elifesciences.org/lax:65774%2Felife-65774-fig4-figsupp3-v2.tif/full/1500,/0/default.jpg
def viral_load(data_registry):
    return np.linspace(
        weibull_min.ppf(
            data_registry.virological_data['covid_overal_vl_data']['parameters']['start'],
            c=data_registry.virological_data['covid_overal_vl_data']['parameters']['shape_factor'],
            scale=data_registry.virological_data['covid_overal_vl_data']['parameters']['scale_factor']
        ),
        weibull_min.ppf(
            data_registry.virological_data['covid_overal_vl_data']['parameters']['stop'],
            c=data_registry.virological_data['covid_overal_vl_data']['parameters']['shape_factor'],
            scale=data_registry.virological_data['covid_overal_vl_data']['parameters']['scale_factor']
        ),
        int(data_registry.virological_data['covid_overal_vl_data']['parameters']['num'])
)
def frequencies_pdf(data_registry):
    return weibull_min.pdf(
        viral_load(data_registry),
        c=data_registry.virological_data['covid_overal_vl_data']['parameters']['shape_factor'],
        scale=data_registry.virological_data['covid_overal_vl_data']['parameters']['scale_factor']
    )
def covid_overal_vl_data(data_registry):
    return LogCustom(
        bounds=(data_registry.virological_data['covid_overal_vl_data']['parameters']['min_bound'], data_registry.virological_data['covid_overal_vl_data']['parameters']['max_bound']),
        function=lambda d: np.interp(
            d,
            viral_load(data_registry),
            frequencies_pdf(data_registry),
            data_registry.virological_data['covid_overal_vl_data']['parameters']['interpolation_fp_left'],
            data_registry.virological_data['covid_overal_vl_data']['parameters']['interpolation_fp_right']
        ),
        max_function=data_registry.virological_data['covid_overal_vl_data']['parameters']['max_function']
    )


# Derived from data in doi.org/10.1016/j.ijid.2020.09.025 and
# https://iosh.com/media/8432/aerosol-infection-risk-hospital-patient-care-full-report.pdf (page 60)
def viable_to_RNA_ratio_distribution():
    return Uniform(0.01, 0.6)


# From discussion with virologists
def infectious_dose_distribution():
    return Uniform(10., 100.)


# From https://doi.org/10.1101/2021.10.14.21264988 and references therein
def virus_distributions(data_registry):
    vd = data_registry.virological_data['virus_distributions']
    return {
        'SARS_CoV_2': mc.SARSCoV2(
            viral_load_in_sputum=evaluate_vl(vd['SARS_CoV_2'], 'viral_load_in_sputum', data_registry),
            infectious_dose=param_evaluation(vd['SARS_CoV_2'], 'infectious_dose'),
            viable_to_RNA_ratio=param_evaluation(vd['SARS_CoV_2'], 'viable_to_RNA_ratio'),
            transmissibility_factor=vd['SARS_CoV_2']['transmissibility_factor']['value'],
            infectiousness_days=vd['SARS_CoV_2']['infectiousness_days']['value'],
        ),
        'SARS_CoV_2_ALPHA': mc.SARSCoV2(
            viral_load_in_sputum=evaluate_vl(vd['SARS_CoV_2_ALPHA'], 'viral_load_in_sputum', data_registry),
            infectious_dose=param_evaluation(vd['SARS_CoV_2_ALPHA'], 'infectious_dose'),
            viable_to_RNA_ratio=param_evaluation(vd['SARS_CoV_2_ALPHA'], 'viable_to_RNA_ratio'),
            transmissibility_factor=vd['SARS_CoV_2_ALPHA']['transmissibility_factor']['value'],
            infectiousness_days=vd['SARS_CoV_2_ALPHA']['infectiousness_days']['value'],
        ),
        'SARS_CoV_2_BETA': mc.SARSCoV2(
            viral_load_in_sputum=evaluate_vl(vd['SARS_CoV_2_BETA'], 'viral_load_in_sputum', data_registry),
            infectious_dose=param_evaluation(vd['SARS_CoV_2_BETA'], 'infectious_dose'),
            viable_to_RNA_ratio=param_evaluation(vd['SARS_CoV_2_BETA'], 'viable_to_RNA_ratio'),
            transmissibility_factor=vd['SARS_CoV_2_BETA']['transmissibility_factor']['value'],
            infectiousness_days=vd['SARS_CoV_2_BETA']['infectiousness_days']['value'],
        ),
        'SARS_CoV_2_GAMMA': mc.SARSCoV2(
            viral_load_in_sputum=evaluate_vl(vd['SARS_CoV_2_GAMMA'], 'viral_load_in_sputum', data_registry),
            infectious_dose=param_evaluation(vd['SARS_CoV_2_GAMMA'], 'infectious_dose'),
            viable_to_RNA_ratio=param_evaluation(vd['SARS_CoV_2_GAMMA'], 'viable_to_RNA_ratio'),
            transmissibility_factor=vd['SARS_CoV_2_GAMMA']['transmissibility_factor']['value'],
            infectiousness_days=vd['SARS_CoV_2_GAMMA']['infectiousness_days']['value'],
        ),
        'SARS_CoV_2_DELTA': mc.SARSCoV2(
            viral_load_in_sputum=evaluate_vl(vd['SARS_CoV_2_DELTA'], 'viral_load_in_sputum', data_registry),
            infectious_dose=param_evaluation(vd['SARS_CoV_2_DELTA'], 'infectious_dose'),
            viable_to_RNA_ratio=param_evaluation(vd['SARS_CoV_2_DELTA'], 'viable_to_RNA_ratio'),
            transmissibility_factor=vd['SARS_CoV_2_DELTA']['transmissibility_factor']['value'],
            infectiousness_days=vd['SARS_CoV_2_DELTA']['infectiousness_days']['value'],
        ),
        'SARS_CoV_2_OMICRON': mc.SARSCoV2(
            viral_load_in_sputum=evaluate_vl(vd['SARS_CoV_2_OMICRON'], 'viral_load_in_sputum', data_registry),
            infectious_dose=param_evaluation(vd['SARS_CoV_2_OMICRON'], 'infectious_dose'),
            viable_to_RNA_ratio=param_evaluation(vd['SARS_CoV_2_OMICRON'], 'viable_to_RNA_ratio'),
            transmissibility_factor=vd['SARS_CoV_2_OMICRON']['transmissibility_factor']['value'],
            infectiousness_days=vd['SARS_CoV_2_OMICRON']['infectiousness_days']['value'],
        ),
    }


# From:
# https://doi.org/10.1080/02786826.2021.1890687
# https://doi.org/10.1016/j.jhin.2013.02.007
# https://doi.org/10.4209/aaqr.2020.08.0531
# https://doi.org/10.1080/02786826.2021.1890687
def mask_distributions(data_registry):
    return {
        'Type I': mc.Mask(
            η_inhale=param_evaluation(
                data_registry.mask_distributions['Type I'], 'η_inhale'),
            η_exhale=param_evaluation(
                data_registry.mask_distributions['Type I'], 'η_exhale')
            if data_registry.mask_distributions['Type I'].get('η_exhale') is not None else None,
            factor_exhale=data_registry.mask_distributions['Type I']['factor_exhale']['value']
        ),
        'FFP2': mc.Mask(
            η_inhale=param_evaluation(
                data_registry.mask_distributions['FFP2'], 'η_inhale'),
            η_exhale=param_evaluation(
                data_registry.mask_distributions['FFP2'], 'η_exhale')
            if data_registry.mask_distributions['FFP2'].get('η_exhale') is not None else None,
            factor_exhale=data_registry.mask_distributions['FFP2']['factor_exhale']['value']
        ),
        'Cloth': mc.Mask(
            η_inhale=param_evaluation(
                data_registry.mask_distributions['Cloth'], 'η_inhale'),
            η_exhale=param_evaluation(
                data_registry.mask_distributions['Cloth'], 'η_exhale')
            if data_registry.mask_distributions['Cloth'].get('η_exhale') is not None else None,
            factor_exhale=data_registry.mask_distributions['Cloth']['factor_exhale']['value']
        ),
    }


def expiration_distribution(
        data_registry,
        BLO_factors,
        d_min=0.1,
        d_max=30.,
):
    """
    Returns an Expiration with an aerosol diameter distribution, defined
    by the BLO factors (a length-3 tuple).
    The total concentration of aerosols, cn, is computed by integrating
    the distribution between 0.1 and 30 microns - these boundaries are
    an historical choice based on previous implementations of the model
    (it limits the influence of the O-mode).
    """
    dscan = np.linspace(d_min, d_max, 3000)
    return mc.Expiration(
        CustomKernel(
            dscan,
            BLOmodel(data_registry, BLO_factors).distribution(dscan),
            kernel_bandwidth=0.1,
        ),
        cn=BLOmodel(data_registry, BLO_factors).integrate(d_min, d_max),
    )


def expiration_BLO_factors(data_registry):
    breathing = data_registry.expiration_particle['expiration_BLO_factors']['Breathing']
    speaking = data_registry.expiration_particle['expiration_BLO_factors']['Speaking']
    singing = data_registry.expiration_particle['expiration_BLO_factors']['Singing']
    shouting = data_registry.expiration_particle['expiration_BLO_factors']['Shouting']
    return {
        'Breathing': (
            param_evaluation(breathing, 'B'),
            param_evaluation(breathing, 'L'),
            param_evaluation(breathing, 'O')
        ),
        'Speaking': (
            param_evaluation(speaking, 'B'),
            param_evaluation(speaking, 'L'),
            param_evaluation(speaking, 'O')
        ),
        'Singing': (
            param_evaluation(singing, 'B'),
            param_evaluation(singing, 'L'),
            param_evaluation(singing, 'O')
        ),
        'Shouting': (
            param_evaluation(shouting, 'B'),
            param_evaluation(shouting, 'L'),
            param_evaluation(shouting, 'O')
        ),
    }


def expiration_distributions(data_registry):
    return {
        exp_type: expiration_distribution(
            data_registry=data_registry,
            BLO_factors=BLO_factors,
            d_min=param_evaluation(data_registry.expiration_particle['particle_size_range']['long_range'], 'minimum_diameter'),
            d_max=param_evaluation(data_registry.expiration_particle['particle_size_range']['long_range'], 'maximum_diameter')
        )
        for exp_type, BLO_factors in expiration_BLO_factors(data_registry).items()
    }


def short_range_expiration_distributions(data_registry):
    return {
        exp_type: expiration_distribution(
            data_registry=data_registry,
            BLO_factors=BLO_factors,
            d_min=param_evaluation(data_registry.expiration_particle['particle_size_range']['short_range'], 'minimum_diameter'),
            d_max=param_evaluation(data_registry.expiration_particle['particle_size_range']['short_range'], 'maximum_diameter')
        )
        for exp_type, BLO_factors in expiration_BLO_factors(data_registry).items()
    }


# Derived from Fig 8 a) "stand-stand" in https://www.mdpi.com/1660-4601/17/4/1445/htm
distances = np.array((0.5, 0.6, 0.7, 0.8, 0.9, 1, 1.1, 1.2,
                     1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2))
frequencies = np.array((0.0598036, 0.0946154, 0.1299152, 0.1064905, 0.1099066, 0.0998209, 0.0845298,
                       0.0479286, 0.0406084, 0.039795, 0.0205997, 0.0152316, 0.0118155, 0.0118155, 0.018485, 0.0205997))
def short_range_distances(data_registry):
    return Custom(
        bounds=(
            param_evaluation(data_registry.short_range_model['conversational_distance'], 'minimum_distance'),
            param_evaluation(data_registry.short_range_model['conversational_distance'], 'maximum_distance')
        ),
        function=lambda x: np.interp(x, distances, frequencies, left=0., right=0.),
        max_function=0.13
    )



import datetime
import functools
import json
from pathlib import Path
import typing

import dateutil.tz
import numpy as np
from scipy.spatial import cKDTree
from timezonefinder import TimezoneFinder


WX_DATA_LOCATION = Path(__file__).absolute().parent
WxStationIdType = str
MonthType = str
# HourlyTempType - 24 temperatures, one for each hour of the day (the average for the given month).
HourlyTempType = typing.List[float]
WxStationRecordType = typing.Tuple[WxStationIdType, str, float, float]


@functools.lru_cache()
def wx_data() -> typing.Dict[WxStationIdType, typing.Dict[MonthType, HourlyTempType]]:
    """
    Load the weather data (temperature in kelvin).

    The data is structured by station location, and for each station location, by month.

    """
    with (WX_DATA_LOCATION / 'global_weather_set.json').open("r") as json_file:
        data = json.load(json_file)

    for station in list(data.keys()):
        for month in list(data[station].keys()):
            if not np.any(np.isnan(data[station][month])):
                data[station][month] = tuple(
                    273.15 + np.array(data[station][month]))
    return data


@functools.lru_cache()
def wx_station_data() -> typing.Dict[WxStationIdType, WxStationRecordType]:
    """
    Return a dictionary of ``station-id: station records``, where station records
    are of the form ``(station-id, station-name, station-latitude, station-longitude)``.

    The stations returned are guaranteed to have valid weather data.

    """
    weather_data = wx_data()
    station_data = {}
    fixed_delimits = [0, 12, 13, 44, 51, 60, 69, 90, 91]
    station_file = WX_DATA_LOCATION / 'hadisd_station_fullinfo_v311_202001p.txt'

    for line in station_file.open('rt'):
        start_end_positions = zip(fixed_delimits[:-1], fixed_delimits[1:])
        split_vals = [line[start:end] for start, end in start_end_positions]
        station_location = (
            split_vals[0], split_vals[2], float(split_vals[3]), float(split_vals[4]),
        )
        # We only consider stations with weather data, don't include the rest.
        if split_vals[0] in weather_data:
            station_data[split_vals[0]] = station_location
    return station_data


@functools.lru_cache()
def _wx_station_kdtree() -> cKDTree:
    """Build a kd-tree of wx station longitude & latitudes (note the coordinate order)"""
    station_data = wx_station_data().values()
    coords = np.array([(stn_record[3], stn_record[2])
                      for stn_record in station_data])
    return cKDTree(coords)


def mean_hourly_temperatures(wx_station: str, month: int) -> HourlyTempType:
    """
    Return the mean monthly temperature for the given weather station and month.

    Returns
    -------

    temperatures: List[24 floats]
        A list containing 24 temperature values, one for each hour, in kelvin.
        Index 0 of the result corresponds to hour 00:00 (UTC), and index 23 (the last) to 23:00 (UTC).

    """
    # Note that the current dataset encodes month number as a string.
    return wx_data()[wx_station][str(month)]


def timezone_at(*, latitude: float, longitude: float) -> datetime.tzinfo:
    """Find a timezone for the given location, or raise."""
    tf = TimezoneFinder()
    tz_name = tf.timezone_at(lat=latitude, lng=longitude)
    tz = dateutil.tz.gettz(tz_name)
    if tz_name is None or tz is None:
        raise ValueError(
            f"Unable to determine the timezone of given location "
            f"(lat={latitude}, lng={longitude})"
        )
    return tz


def refine_hourly_data(source_times, hourly_data, npts):
    """
    Given times (in hours), where each data point is on the hour,
    interpolate the data to mid-point of the returned boundaries.

    For example:

    >>> time_bounds, data = refine_hourly_data(list(range(24)), list(range(24)), 24)
    >>> len(time_bounds), len(data)
    (25, 24)
    >>> time_bounds
    array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9., 10., 11., 12.,
           13., 14., 15., 16., 17., 18., 19., 20., 21., 22., 23., 24.])
    >>> data
    array([ 0.5,  1.5,  2.5,  3.5,  4.5,  5.5,  6.5,  7.5,  8.5,  9.5, 10.5,
           11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5, 18.5, 19.5, 20.5, 21.5,
           22.5, 11.5])

    The source times need not be monotonic, which allows for data to be
    time-offset shifted. For example:

    >>> time_bounds, data = refine_hourly_data(
    ...     list(range(6, 28)) + [4, 5], list(range(24)), 24)
    >>> data
    array([18.5, 19.5, 20.5, 21.5, 22.5, 11.5,  0.5,  1.5,  2.5,  3.5,  4.5,
            5.5,  6.5,  7.5,  8.5,  9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5,
           16.5, 17.5])

    """
    target_time_boundaries, step = np.linspace(
        0, 24, npts + 1, retstep=True, endpoint=True,
    )
    target_times = target_time_boundaries[:-1] + step / 2
    data = np.interp(target_times, np.array(source_times), hourly_data, period=24)
    return target_time_boundaries, data


def nearest_wx_station(*, longitude: float, latitude: float) -> WxStationRecordType:
    """
    Given a latitude & longitude, return the nearest station with valid weather data.

    """
    ktree = _wx_station_kdtree()
    station_data = list(wx_station_data().values())
    dd, ii = ktree.query((longitude, latitude), k=[1])
    return station_data[ii[0]]



import functools


def method_cache(fn):
    """
    A decorator for instance based caching.

    Unlike lru_cache / memoization, this allows us to not have to have the
    instance itself be hashable - only the arguments must be so.

    The cache is stored as a dictionary in a private attribute on the instance
    with the name ``_cache_{func_name}``.

    """
    cache_name = f'_cache_{fn.__name__}'

    @functools.wraps(fn)
    def cached_method(self, *args, **kwargs):
        cache = getattr(self, cache_name, None)
        if cache is None:
            cache = {}
            object.__setattr__(self, cache_name, cache)
        cache_key = hash(args + tuple(kwargs.items()))
        if cache_key not in cache:
            cache[cache_key] = fn(self, *args, **kwargs)
        return cache[cache_key]
    return cached_method
